/*****************************************************************************/
/*
                                 Config.c

This module performs three functions.

  1.  It reads the HTTPD configuration file, converting the directives
      into internal representations.

  2.  Provides functions, accessed by other modules, returning the
      original directive information as appropriate.

  3.  Provides HTML form-based configuration and reports.

Configuration can only be performed at server startup.  To reset the 
configuration the server must be restarted.

WASD HTTPd Configuration is a relevant subset of the CERN configuration 
directives (sort-of, well it started that way anyhow!)  Although there are
some  differences (beware!) the CERN documentation can be consulted for a
general discussion of the purpose and functioning some of these directives. 

The file formats may be IN ONE OF TWO possible formats (not mixed).  The
older format contains one directive at the start of a line, followed by a
mandatory parameter.  The newer format (in part to support forms-based
configuration) has the directive contained within square brackets (e.g.
"[accept]") followed on the same line, or by one or more lines containing
comments, containing parameter information, or empty.  The directive applies
to all lines following up until another directive (square bracket in column
one) is encountered.

Directives are:

  o  Accept             <host/domain>
  o  ActivityDays       <integer>
  o  AddType            <suffix> <content-type> -|</script> [<description>]
  o  AddIcon            <URL> <alternative-text> <content-type-template>
  o  AddDirIcon         <URL> <alternative-text>
  o  AddBlankIcon       <URL> <alternative-text>
  o  AddUnknownIcon     <URL> <alternative-text>
  o  AddParentIcon      <URL> <alternative-text>
  o  AuthBasic          DISABLED | ENABLED
  o  AuthCacheMinutes   <integer>
  o  AuthDigest         DISABLED | ENABLED
  o  AuthDigestGetLife  <integer>
  o  AuthDigestPutLife  <integer>
  o  AuthFailureLimit   <integer>
  o  AuthRevalidateUserMinutes   <integer>
  o  BufferSizeDclCgiPlusIn      <integer>
  o  BufferSizeDclCommand        <integer>
  o  BufferSizeDclOutput         <integer>
  o  BufferSizeNetRead           <integer>
  o  BufferSizeNetWrite          <integer>
  o  Busy               <integer>
  o  Cache              DISABLED | ENABLED
  o  CacheChunkKBytes   <integer>
  o  CacheEntriesMax    <integer>
  o  CacheFileKBytesMax <integer>
  o  CacheFrequentHits      <integer>
  o  CacheFrequentSeconds   <integer>
  o  CacheHashTableEntries  <integer>
  o  CacheTotalKBytesMax    <integer>
  o  CacheValidateSeconds   <integer>
  o  CharsetDefault     <string>
  o  CommentedInfo      DISABLED | ENABLED
  o  DclCgiPlusLifeTime <integer>
  o  DclHardLimit       <integer>
  o  DclScriptRunTime   <string>
  o  DclSoftLimit       <integer>
  o  DclSpawnAuthPriv   DISABLED | ENABLED
  o  DclFullRequest     DISABLED | ENABLED
  o  DclZombieLifeTime  <integer>
  o  DECnetReuseLifeTime    <integer>
  o  DECnetConnectListMax   <integer>
  o  DirAccess          DISABLED | ENABLED | SELECTIVE
  o  DirDescription     <integer>
  o  DirLayout          <string>
  o  DirOwner           DISABLED | ENABLED
  o  DirPreExpired      DISABLED | ENABLED
  o  DirReadMe          DISABLED | TOP | BOTTOM
  o  DirReadMeFile      <file.suffix>
  o  DirNoImpliedWildcard   DISABLED | ENABLED
  o  DirNoPrivIgnore    DISABLED | ENABLED
  o  DirWildcard        DISABLED | ENABLED
  o  DNSLookup          DISABLED | ENABLED
  o  ErrorRecommend     DISABLED | ENABLED
  o  ErrorReportPath    <path-to-report>
  o  ErrorSourceInfo    DISABLED | ENABLED
  o  Logging            DISABLED | ENABLED
  o  LogExcludeHosts    <string>
  o  LogFile            <string>
  o  LogFormat          <string>
  o  LogNaming          <string>
  o  LogPeriod          <string>
  o  LogPerService      DISABLED | ENABLED
  o  MapUserNameCacheEntries          <integer>
  o  Monitor            DISABLED | ENABLED
  o  Port               <integer>
  o  ProxyAddForwardedBy              DISABLED | ENABLED
  o  ProxyCache                       DISABLED | ENABLED
  o  ProxyCacheDeviceCheckMinutes     <integer>
  o  ProxyCacheFileKBytesMax          <integer>
  o  ProxyCacheMaxPercent             <integer>
  o  ProxyCachePurgeList              <string>
  o  ProxyCacheReducePercent          <integer>
  o  ProxyCacheReloadList             <string>
  o  ProxyCacheRoutineHourOfDay       <integer>
  o  ProxyHostCachePurgeHours         <integer>
  o  ProxyHostLookupRetryCount        <integer>
  o  ProxyReportCacheLog              DISABLED | ENABLED
  o  ProxyReportLog                   DISABLED | ENABLED
  o  ProxyServing                     DISABLED | ENABLED
  o  PutMaxKbytes       <integer>
  o  Reject             <host/domain>
  o  RequestHistory     <integer>
  o  SearchScript       <path-to-search-script>
  o  Service            <string>
  o  ServiceNotFoundURL <string>
  o  Ssi | Shtml                      DISABLED | ENABLED
  o  SsiAccesses | ShtmlAccesses      DISABLED | ENABLED
  o  SsiExec | ShtmlExec              DISABLED | ENABLED
  o  StreamLf           <integer>
  o  TimeoutInput       <integer>
  o  TimeoutKeepAlive   <integer>
  o  TimeoutNoProgress  <integer>
  o  TimeoutOutput      <integer>
  o  Welcome            <file.suffix>


VERSION HISTORY
---------------
30-OCT-1999  MGD  include IP address for services in ConfigReportNow(),
                  'Buffer...', 'MapUserNameCacheEntries', 'ServiceNotFoundURL',
                  'Timeout...' directives (with backward compatibility),
                  increase buffer space for ConfigCommaList()
03-MAR-1999  MGD  improve [AddType] and [AddIcon] directive checking
18-JAN-1999  MGD  [Proxy...] configuration parameters
                  bugfix; ignore empty lines (even with spaces in them!)
17-OCT-1998  MGD  [ErrorReportPath] added,
                  [Search] changed to [SearchScript],
                  [StreamLfPaths] obsoleted in favour of SET mapping rule
29-SEP-1998  MGD  [AddType] now allows "text/html; charset=ISO8859-5",
                  [CharsetDefault] text and server character set
27-AUG-1998  MGD  generic ConfigSetCommaList(),
                  add [AuthRevalidateUserMinutes], [LogExcludeHosts],
                      [DECnetReuseLifeTime], [DECnetConnectListMax],
                      [StreamLfPaths]
08-AUG-1998  MGD  add [DirNoImpliedWildcard]
17-JUN-1998  MGD  add [LogNaming], [LogPerService]
28-FEB-1998  MGD  improved icon handling efficiency
07-FEB-1998  MGD  v5.0 hash table for content-type suffix and icon searches,
                  default content type (for unknowns) configurable,
                  HTTP protocol name in [service] (for SSL),
                  [DirDescription] changed from boolean to integer
05-OCT-1997  MGD  cache parameters,
                  DCL script run-time parameter,
                  additional logging parameters
09-AUG-1997  MGD  v4.4 retired "ErrorInfo", "ErrorSysAdmin" (now in messages)
                  and AuthVMS (now /SYSUAF qualifier);  introduced
                  "DirDescription", "DirNoPrivIgnore", "ErrorSourceInfo",
                  bugfix; memory leak in file-based report and revise
01-AUG-1997  MGD  v4.3 activity number of days, DCL supply full request,
                  bugfix; annoying repeated instances of "unknown" content type
01-JUN-1997  MGD  v4.2 required additional DCL/scripting parameters
01-FEB-1997  MGD  HTTPd version 4;
                  major changes for form-based config;
                  added error message URLs
01-OCT-1996  MGD  HTML configuration report
01-AUG-1996  MGD  StreamLF conversion directive;
                  AuthBasic and AuthDigest directives;
12-APR-1996  MGD  file record/binary now determined by record format
                  (has removed the need of the text/binary, etc., in "AddType")
01-DEC-1995  MGD  HTTPd version 3
27-SEP-1995  MGD  added auto-scripting funtionality
07-AUG-1995  MGD  added support for VMS-style directory listing
16-JUN-1995  MGD  file contents description in 'AddType' (also see DIR.C)
20-DEC-1994  MGD  developed for multi-threaded HTTP daemon
*/
/*****************************************************************************/

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>

/* VMS related header files */
#include <rmsdef.h>
#include <rms.h>
#include <ssdef.h>
#include <stsdef.h>

/* application related header */
#include "wasd.h"

#define WASD_MODULE "CONFIG"

/******************/
/* global storage */
/******************/

struct ConfigStruct  Config;

char  ConfigContentTypeBlank [] = "x-internal/blank",
      ConfigContentTypeDir [] = "x-internal/directory",
      ConfigContentTypeIsMap [] = "text/x-ismap",
      ConfigContentTypeMenu [] = "text/x-menu",
      ConfigContentTypeParent [] = "x-internal/parent",
      ConfigContentTypeSsi [] = "text/x-shtml",
      ConfigContentTypeUnknown [] = "x-internal/unknown",
      ConfigDefaultFileContentType [] = "application/octet-stream",
      ConfigStringOverflow [] = "*ERROR* string overflow";


char  *ConfigBlankIconPtr,
      *ConfigDirIconPtr,
      *ConfigParentIconPtr,
      *ConfigUnknownIconPtr;

/********************/
/* external storage */
/********************/

#ifdef DBUG
extern boolean Debug;
#else
#define Debug 0 
#endif

extern int  DclCgiPlusInSize,
            DclSysCommandSize,
            DclSysOutputSize,
            MapUrlUserNameCacheEntries,
            NetReadBufferSize,
            OutputBufferSize,
            ServiceCount;

extern char  *CliAcceptHostsPtr,
             *CliRejectHostsPtr;

extern char  ServerHostPort[],
             SoftwareID[],
             TimeGmtString[],
             Utility[];

extern struct MsgStruct  Msgs;

/*****************************************************************************/
/*
Open the configuration file defined by the logical name HTTPD$CONFIG.  Read 
each line, passing to the appropriate interpretation function according to the 
directive on the line.  Close the file.  Function is either called before
AST-driven  processing or called at AST delivery level which makes processing
atomic.  Allow 32 extra characters on line buffer and parameter allocated
memory for concatenating additional string.
*/

Configure (struct ConfigStruct *cfptr)

{
#  define PARAMETER_ELBOW_ROOM 32

   register char  *cptr, *sptr, *zptr;

   boolean  DebugBuffer,
            BracketDirectives,
            EquateDirectives;
   int  status,
        LineLength;
   void  *RequestPtrBuffer;
   char  ExpandedFileName [256],
         Line [256 + PARAMETER_ELBOW_ROOM];
   struct FAB  ConfigFileFab;
   struct NAM  ConfigFileNam;
   struct RAB  ConfigFileRab;
   struct XABDAT  ConfigFileXabDat;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "Configure()\n");

#ifdef DBUG
   /* not interested anymore in seeing debug information for config load! */
   DebugBuffer = Debug;
   Debug = false;
#endif

   RequestPtrBuffer = cfptr->RequestPtr;
   memset (cfptr, 0, sizeof(struct ConfigStruct));
   cfptr->RequestPtr = RequestPtrBuffer;

   BracketDirectives = EquateDirectives = false;
   sys$gettim (&cfptr->LoadBinTime);

   /************************/
   /* open the config file */
   /************************/

   if (Debug) fprintf (stdout, "CONFIG_FILE_NAME |%s|\n", CONFIG_FILE_NAME);

   ConfigFileFab = cc$rms_fab;
   ConfigFileFab.fab$b_fac = FAB$M_GET;
   ConfigFileFab.fab$l_fna = CONFIG_FILE_NAME;
   ConfigFileFab.fab$b_fns = strlen(CONFIG_FILE_NAME);
   ConfigFileFab.fab$l_nam = &ConfigFileNam;
   ConfigFileFab.fab$b_shr = FAB$M_SHRGET;
   ConfigFileFab.fab$l_xab = &ConfigFileXabDat;

   ConfigFileNam = cc$rms_nam;
   ConfigFileNam.nam$l_esa = ExpandedFileName;
   ConfigFileNam.nam$b_ess = sizeof(ExpandedFileName)-1;
   ConfigFileNam.nam$l_rsa = cfptr->LoadFileName;
   ConfigFileNam.nam$b_rss = sizeof(cfptr->LoadFileName)-1;

   /* initialize the date extended attribute block */
   ConfigFileXabDat = cc$rms_xabdat;

   /* turn on SYSPRV to allow access to possibly protected file */
   EnableSysPrv();
   status = sys$open (&ConfigFileFab, 0, 0);
   DisableSysPrv();

   /* status from sys$open() */
   if (VMSnok (status))
   {
      if (Debug) fprintf (stdout, "sys$open() %%X%08.08X\n", status);
      if (cfptr->RequestPtr)
      {
         ConfigReportProblem (cfptr, NULL, status);
         return;
      }
      else
         ErrorExitVmsStatus (status, CONFIG_FILE_NAME, FI_LI);
   }

   cfptr->LoadFileName[ConfigFileNam.nam$b_rsl] = '\0';
   if (Debug)
      fprintf (stdout, "cfptr->LoadFileName |%s|\n", cfptr->LoadFileName);

   /* get the configuration file revision date and time */
   memcpy (&cfptr->RevBinTime,
           &ConfigFileXabDat.xab$q_rdt,
           sizeof(cfptr->RevBinTime));

   /* record access block */
   ConfigFileRab = cc$rms_rab;
   ConfigFileRab.rab$l_fab = &ConfigFileFab;
   /* try and optimise sequential read (next 3 lines) */
   ConfigFileRab.rab$b_mbc = 6;
   ConfigFileRab.rab$b_mbf = 2;
   ConfigFileRab.rab$l_rop = RAB$M_RAH;
   ConfigFileRab.rab$l_ubf = Line;
   ConfigFileRab.rab$w_usz = sizeof(Line)-PARAMETER_ELBOW_ROOM-1;

   if (VMSnok (status = sys$connect (&ConfigFileRab, 0, 0)))
   {
      sys$close (&ConfigFileFab, 0, 0);
      if (Debug) fprintf (stdout, "sys$connect() %%X%08.08X\n", status);
      if (cfptr->RequestPtr)
      {
         ConfigReportProblem (cfptr, NULL, status);
         return;
      }
      else
         ErrorExitVmsStatus (status, CONFIG_FILE_NAME, FI_LI);
   }

   /************************/
   /* read the config file */
   /************************/

   cfptr->AuthCacheMinutes =
      cfptr->AuthRevalidateUserMinutes = 60;
   cfptr->AuthFailureLimit = 10;
   cfptr->DirPreExpired = PRE_EXPIRE_INDEX_OF;
   strcpy (cfptr->DefaultDirLayout, DEFAULT_DIR_LAYOUT);
   cfptr->AcceptHostsPtr =
      cfptr->ContentTypeListHeadPtr =
      cfptr->IconListHeadPtr =
      cfptr->LogExcludeHostsPtr =
      cfptr->RejectHostsPtr = NULL;
   cfptr->ContentTypeStructOverhead = sizeof(struct ContentTypeStruct);
   cfptr->IconStructOverhead = sizeof(struct IconStruct);
   cfptr->ContentTypeDefaultPtr = "";

   while (VMSok (status = sys$get (&ConfigFileRab, 0, 0)))
   {
      cfptr->LineNumber++;
      Line[LineLength = ConfigFileRab.rab$w_rsz] = '\0';
      if (Debug) fprintf (stdout, "Line %d |%s|\n", cfptr->LineNumber, Line);

      for (cptr = Line; ISLWS(*cptr); cptr++);

      /* if a blank or comment line, then ignore */
      if (!*cptr || *cptr == '!' || *cptr == '#') continue;

      if (Line[0] == '[')
      {
         if (EquateDirectives)
         {
            ConfigReportProblem (cfptr, "mixed directive formats", 0);
            break;
         }
         BracketDirectives = true;
         cptr++;
         zptr = (sptr = cfptr->DirectiveName) + sizeof(cfptr->DirectiveName);
         while (*cptr && *cptr != ']' && sptr < zptr) *sptr++ = *cptr++;
         if (sptr >= zptr)
         {
            ConfigReportProblem (cfptr, "missing closing bracket?", 0);
            continue;
         }
         *sptr = '\0';
         if (Debug)
            fprintf (stdout, "DirectiveName |%s|\n", cfptr->DirectiveName);
         if (*cptr) cptr++;
         while (*cptr && ISLWS(*cptr)) cptr++;
      }
      else
      if (!BracketDirectives)
      {
         EquateDirectives = true;
         zptr = (sptr = cfptr->DirectiveName) + sizeof(cfptr->DirectiveName);
         while (*cptr && !ISLWS(*cptr) && *cptr != '=' && sptr < zptr)
             *sptr++ = *cptr++;
         if (sptr >= zptr)
         {
            ConfigReportProblem (cfptr, "missing parameter");
            continue;
         }
         *sptr = '\0';
         while (*cptr && ISLWS(*cptr)) cptr++;
         if (!*cptr)
         {
            ConfigReportProblem (cfptr, "missing parameter", 0);
            continue;
         }
      }

      if (BracketDirectives && !*cptr) continue;

      if (strsame (cfptr->DirectiveName, "Accept", -1))
         ConfigSetCommaList (cfptr, cptr, &cfptr->AcceptHostsPtr,
                                          &cfptr->AcceptHostsLength);
      else
      if (strsame (cfptr->DirectiveName, "ActivityDays", -1))
         cfptr->ActivityNumberOfDays = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "AddType", -1))
         ConfigAddType (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "AddIcon", -1))
         ConfigAddIcon (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "AddBlankIcon", -1))
      {
         /* use up the PARAMETER_ELBOW_ROOM :^) */
         for (sptr = cptr; *sptr; sptr++);
         *sptr++ = ' ';
         strcpy (sptr, ConfigContentTypeBlank);
         ConfigAddIcon (cfptr, cptr);
      }
      else
      if (strsame (cfptr->DirectiveName, "AddDirIcon", -1))
      {
         /* use up the PARAMETER_ELBOW_ROOM :^) */
         for (sptr = cptr; *sptr; sptr++);
         *sptr++ = ' ';
         strcpy (sptr, ConfigContentTypeDir);
         ConfigAddIcon (cfptr, cptr);
      }
      else
      if (strsame (cfptr->DirectiveName, "AddParentIcon", -1))
      {
         /* use up the PARAMETER_ELBOW_ROOM :^) */
         for (sptr = cptr; *sptr; sptr++);
         *sptr++ = ' ';
         strcpy (sptr, ConfigContentTypeParent);
         ConfigAddIcon (cfptr, cptr);
      }
      else
      if (strsame (cfptr->DirectiveName, "AddUnknownIcon", -1))
      {
         /* use up the PARAMETER_ELBOW_ROOM :^) */
         for (sptr = cptr; *sptr; sptr++);
         *sptr++ = ' ';
         strcpy (sptr, ConfigContentTypeUnknown);
         ConfigAddIcon (cfptr, cptr);
      }
      else
      if (strsame (cfptr->DirectiveName, "AuthBasic", -1))
         cfptr->AuthBasicEnabled = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "AuthCacheMinutes", -1))
         cfptr->AuthCacheMinutes = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "AuthDigestGetLife", -1))
         cfptr->AuthDigestNonceGetLifeTime = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "AuthDigestPutLife", -1))
         cfptr->AuthDigestNoncePutLifeTime = ConfigSetInteger (cfptr, cptr);
      else
      /* "AuthDigest" MUST come after other AuthDigests for obvious reasons */
      if (strsame (cfptr->DirectiveName, "AuthDigest", -1))
         cfptr->AuthDigestEnabled = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "AuthFailureLimit", -1))
         cfptr->AuthFailureLimit = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "AuthRevalidateUserMinutes", -1))
         cfptr->AuthRevalidateUserMinutes = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "BufferSizeDclCommand", -1))
         cfptr->BufferSizeDclCommand = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "BufferSizeDclOutput", -1))
         cfptr->BufferSizeDclOutput = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "BufferSizeDclCgiPlusIn", -1))
         cfptr->BufferSizeDclCgiPlusIn = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "BufferSizeNetRead", -1))
         cfptr->BufferSizeNetRead = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "BufferSizeNetWrite", -1))
         cfptr->BufferSizeNetWrite = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "Busy", -1))
         cfptr->Busy = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "CacheChunkKBytes", -1))
         cfptr->CacheChunkKBytes = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "CacheEntriesMax", -1))
         cfptr->CacheEntriesMax = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "CacheFileKBytesMax", -1))
         cfptr->CacheFileKBytesMax = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "CacheFrequentHits", -1))
         cfptr->CacheFrequentHits = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "CacheFrequentSeconds", -1))
         cfptr->CacheFrequentSeconds = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "CacheHashTableEntries", -1))
         cfptr->CacheHashTableEntries = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "CacheTotalKBytesMax", -1))
         cfptr->CacheTotalKBytesMax = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "CacheValidateSeconds", -1))
         cfptr->CacheValidateSeconds = ConfigSetInteger (cfptr, cptr);
      else
      /* "cache" must follow all the others for the obvious reason */
      if (strsame (cfptr->DirectiveName, "Cache", -1))
         cfptr->CacheEnabled = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "CharsetDefault", -1))
         ConfigSetString (cfptr, cptr, cfptr->CharsetDefault,
                                       sizeof(cfptr->CharsetDefault));
      else
      if (strsame (cfptr->DirectiveName, "CommentedInfo", -1))
         cfptr->IncludeCommentedInfo = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DclHardLimit", -1))
         cfptr->DclSubprocessHardLimit = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DclScriptRunTime", -1))
         ConfigSetScriptRunTime (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DclSoftLimit", -1))
         cfptr->DclSubprocessSoftLimit = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DclCgiPlusLifeTime", -1))
         cfptr->DclCgiPlusLifeTime = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DclZombieLifeTime", -1))
         cfptr->DclZombieLifeTime = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DclSpawnAuthPriv", -1))
         cfptr->DclSpawnAuthPriv = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DclFullRequest", -1))
         cfptr->DclFullRequest = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DECnetReuseLifeTime", -1))
         cfptr->DECnetReuseLifeTime = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DECnetConnectListMax", -1))
         cfptr->DECnetConnectListMax = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DirAccess", -1))
         ConfigSetDirAccess (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DirDescription", -1))
      {
         /*
            [DirDescription] boolean deprecated in v5.0, noow using
            integer [DirDescriptionLines], provide fallback behaviour.
         */
         if (ConfigSetBoolean (cfptr, cptr))
            cfptr->DirDescriptionLines = 30;
         else
            cfptr->DirDescriptionLines = 0;
      }
      else
      if (strsame (cfptr->DirectiveName, "DirDescriptionLines", -1))
         cfptr->DirDescriptionLines = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DirLayout", -1))
         ConfigSetString (cfptr, cptr,
                          cfptr->DefaultDirLayout,
                          sizeof(cfptr->DefaultDirLayout));
      else
      if (strsame (cfptr->DirectiveName, "DirOwner", -1))
         cfptr->DirOwnerEnabled = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DirPreExpired", -1))
         cfptr->DirPreExpired = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DirNoImpliedWildcard", -1))
         cfptr->DirNoImpliedWildcard = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DirNoPrivIgnore", -1))
         cfptr->DirNoPrivIgnore = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DirReadMeFile", -1))
         ConfigSetDirReadMeFile (cfptr, cptr);
      else
      /* "ReadMe" MUST come after the other ReadMes for obvious reasons */
      if (strsame (cfptr->DirectiveName, "DirReadMe", -1))
         ConfigSetDirReadMe (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DirWildcard", -1))
         cfptr->DirWildcardEnabled = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "DNSLookup", -1))
         cfptr->DnsLookup = ConfigSetBoolean (cfptr, cptr);
      else
      /* "Recommend" for backward compatibility */
      if (strsame (cfptr->DirectiveName, "ErrorRecommend", -1) ||
          strsame (cfptr->DirectiveName, "Recommend", -1))
         cfptr->ErrorRecommend = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "ErrorSourceInfo", -1))
         cfptr->ErrorSourceInfo = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "ErrorReportPath", -1))
      {
         ConfigSetString (cfptr, cptr,
                          cfptr->ErrorReportPath,
                          sizeof(cfptr->ErrorReportPath));
         cfptr->ErrorReportPathLength = strlen(cfptr->ErrorReportPath);
      }
      else
      if (strsame (cfptr->DirectiveName, "Logging", -1))
         cfptr->LoggingEnabled = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "LogExcludeHosts", -1))
         ConfigSetCommaList (cfptr, cptr, &cfptr->LogExcludeHostsPtr,
                                          &cfptr->LogExcludeHostsLength);
      else
      if (strsame (cfptr->DirectiveName, "LogFile", -1))
         ConfigSetString (cfptr, cptr,
                          cfptr->LogFileName, sizeof(cfptr->LogFileName));
      else
      if (strsame (cfptr->DirectiveName, "LogFormat", -1))
         ConfigSetString (cfptr, cptr,
                          cfptr->LogFormat, sizeof(cfptr->LogFormat));
      else
      if (strsame (cfptr->DirectiveName, "LogNaming", -1))
         ConfigSetString (cfptr, cptr,
                          cfptr->LogNaming, sizeof(cfptr->LogNaming));
      else
      if (strsame (cfptr->DirectiveName, "LogPeriod", -1))
         ConfigSetString (cfptr, cptr,
                          cfptr->LogPeriod, sizeof(cfptr->LogPeriod));
      else
      if (strsame (cfptr->DirectiveName, "LogPerService", -1))
         cfptr->LogPerService = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "MapUserNameCacheEntries", -1))
         cfptr->MapUserNameCacheEntries = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "Monitor", -1))
         cfptr->MonitorEnabled = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "ProxyAddForwardedBy", -1))
         cfptr->ProxyAddForwardedByEnabled = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "ProxyCache", -1))
         cfptr->ProxyCacheEnabled = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "ProxyCacheDeviceCheckMinutes", -1))
         cfptr->ProxyCacheDeviceCheckMinutes = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "ProxyCacheDeviceMaxPercent", -1))
         cfptr->ProxyCacheDeviceMaxPercent = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "ProxyCacheDevicePurgePercent", -1))
         cfptr->ProxyCacheDevicePurgePercent = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "ProxyCacheFileKBytesMax", -1))
         cfptr->ProxyCacheFileKBytesMax = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "ProxyCacheRoutineHourOfDay", -1))
         cfptr->ProxyCacheRoutineHourOfDay = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "ProxyCachePurgeList", -1))
         ConfigSetString (cfptr, cptr,
                          cfptr->ProxyCachePurgeList,
                          sizeof(cfptr->ProxyCachePurgeList));
      else
      if (strsame (cfptr->DirectiveName, "ProxyCacheReloadList", -1))
         ConfigSetString (cfptr, cptr,
                          cfptr->ProxyCacheReloadList,
                          sizeof(cfptr->ProxyCacheReloadList));
      else
      if (strsame (cfptr->DirectiveName, "ProxyHostCachePurgeHours", -1))
         cfptr->ProxyHostCachePurgeHours = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "ProxyHostLookupRetryCount", -1))
         cfptr->ProxyHostLookupRetryCount = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "ProxyReportCacheLog", -1))
         cfptr->ProxyReportCacheLog = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "ProxyReportLog", -1))
         cfptr->ProxyReportLog = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "ProxyServing", -1))
         cfptr->ProxyServingEnabled = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "Port", -1))
         cfptr->ServerPort = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "PutMaxKbytes", -1))
         cfptr->PutMaxKbytes = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "PutVersionLimit", -1))
         cfptr->PutVersionLimit = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "Reject", -1))
         ConfigSetCommaList (cfptr, cptr, &cfptr->RejectHostsPtr,
                                          &cfptr->RejectHostsLength);
      else
      if (strsame (cfptr->DirectiveName, "RequestHistory", -1))
         cfptr->RequestHistory = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "Search", -1) ||
          strsame (cfptr->DirectiveName, "SearchScript", -1))
      {
         ConfigSetString (cfptr, cptr,
                          cfptr->KeywordSearchScript,
                          sizeof(cfptr->KeywordSearchScript));
         cfptr->KeywordSearchScriptLength = strlen(cfptr->KeywordSearchScript);
      }
      else
      if (strsame (cfptr->DirectiveName, "Service", -1))
         ConfigSetCommaList (cfptr, cptr, &cfptr->ServicePtr,
                                          &cfptr->ServiceLength);
      else
      /* "ShtmlAccesses" for backward compatibility */
      if (strsame (cfptr->DirectiveName, "SSIAccesses", -1) ||
          strsame (cfptr->DirectiveName, "ShtmlAccesses", -1))
         cfptr->SsiAccessesEnabled = ConfigSetBoolean (cfptr, cptr);
      else
      /* "ShtmlExec" for backward compatibility */
      if (strsame (cfptr->DirectiveName, "SSIExec", -1) ||
          strsame (cfptr->DirectiveName, "ShtmlExec", -1))
         cfptr->SsiExecEnabled = ConfigSetBoolean (cfptr, cptr);
      else
      /* "SSI" MUST come after the other SSIs for obvious reasons */
      /* "Shtml" for backward compatibility */
      if (strsame (cfptr->DirectiveName, "SSI", -1) ||
          strsame (cfptr->DirectiveName, "Shtml", -1))
         cfptr->SsiEnabled = ConfigSetBoolean (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "StreamLF", -1))
         cfptr->StreamLfConversionMaxKbytes =
            ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "TimeoutInput", -1) ||
          strsame (cfptr->DirectiveName, "InputTimeout", -1))
         cfptr->TimeoutMinutesInput = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "TimeoutKeepAlive", -1) ||
          strsame (cfptr->DirectiveName, "KeepAliveTimeout", -1))
         cfptr->TimeoutSecondsKeepAlive = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "TimeoutNoProgress", -1))
         cfptr->TimeoutMinutesNoProgress = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "TimeoutOutput", -1) ||
          strsame (cfptr->DirectiveName, "OutputTimeout", -1))
         cfptr->TimeoutMinutesOutput = ConfigSetInteger (cfptr, cptr);
      else
      if (strsame (cfptr->DirectiveName, "ServiceNotFoundURL", -1))
         ConfigSetString (cfptr, cptr,
                          cfptr->ServiceNotFoundUrl,
                          sizeof(cfptr->ServiceNotFoundUrl));
      else
      if (strsame (cfptr->DirectiveName, "Welcome", -1))
         ConfigSetWelcome (cfptr, cptr);
      else
         ConfigReportProblem (cfptr, "?", 0);
   }

   if (status == RMS$_EOF) status = SS$_NORMAL;

   /*************************/
   /* close the config file */
   /*************************/

   sys$close (&ConfigFileFab, 0, 0); 

   if (!cfptr->PutMaxKbytes)
      cfptr->PutMaxKbytes = DEFAULT_PUT_KBYTES_MAX;
   if (!cfptr->PutVersionLimit)
      cfptr->PutVersionLimit = DEFAULT_PUT_VERSION_LIMIT;

   if (CliAcceptHostsPtr != NULL)
      Config.AcceptHostsPtr = CliAcceptHostsPtr;

   if (CliRejectHostsPtr != NULL)
      Config.RejectHostsPtr = CliRejectHostsPtr;

   ConfigContentTypeIcon (cfptr);

   if (VMSnok (status))
   {
      if (cfptr->RequestPtr)
      {
         ConfigReportProblem (cfptr, NULL, status);
         return;
      }
      else
         ErrorExitVmsStatus (status, CONFIG_FILE_NAME, FI_LI);
   }

#ifdef DBUG
   Debug = DebugBuffer;
#endif

   return (status);
}

/*****************************************************************************/
/*
Return true or false values equivalent to the string pointed to by 'pptr'.
*/

boolean ConfigSetBoolean
(
struct ConfigStruct *cfptr,
char *pptr
)
{
   /*********/
   /* begin */
   /*********/

   if (strsame (pptr, "YES", 3) ||
       strsame (pptr, "ON", 2) ||
       strsame (pptr, "ENABLED", 7))
      return (true);

   if (strsame (pptr, "NO", 2) ||
       strsame (pptr, "OFF", 3) ||
       strsame (pptr, "DISABLED", 8))
      return (false);

   ConfigReportProblem (cfptr, NULL, 0);
   return (false);
}

/*****************************************************************************/
/*
Free all memory allocated for a configuration.  Only used after file-based
configuration report or revision.
*/

ConfigFree (struct ConfigStruct *cfptr)

{
   int  Count;
   char  *cptr;
   struct ContentTypeStruct  *clptr;
   struct IconStruct  *ilptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigFree()\n");

   clptr = cfptr->ContentTypeListHeadPtr;
   while (clptr != NULL)
   {
      cptr = (char*)clptr;
      clptr = clptr->NextPtr;
      VmFree (cptr);
   }

   ilptr = cfptr->IconListHeadPtr;
   while (ilptr != NULL)
   {
      cptr = (char*)ilptr;
      ilptr = ilptr->NextPtr;
      VmFree (cptr);
   }

   for (Count = 0; Count < cfptr->HomePageCount; Count++)
      if (cfptr->HomePageArray[Count] != NULL)
         VmFree (cfptr->HomePageArray[Count]);

   for (Count = 0; Count < cfptr->ScriptRunTimeCount; Count++)
      if (cfptr->ScriptRunTimeArray[Count] != NULL)
         VmFree (cfptr->ScriptRunTimeArray[Count]);

   if (cfptr->AcceptHostsPtr != NULL) VmFree (cfptr->AcceptHostsPtr);
   if (cfptr->LogExcludeHostsPtr != NULL) VmFree (cfptr->LogExcludeHostsPtr);
   if (cfptr->ProblemReportPtr != NULL) VmFree (cfptr->ProblemReportPtr);
   if (cfptr->RejectHostsPtr != NULL) VmFree (cfptr->RejectHostsPtr);
}

/*****************************************************************************/
/*
Return an integer equivalent to the numeric string pointed to by 'pptr'.
*/

int ConfigSetInteger
(
struct ConfigStruct *cfptr,
char *pptr
)
{
   /*********/
   /* begin */
   /*********/

   if (isdigit(*pptr)) return (atoi(pptr));
   ConfigReportProblem (cfptr, NULL, 0);
   return (-1);
}

/*****************************************************************************/
/*
Copy a simple string configuration parameter trimming leadin and trailing
white-space.
*/

ConfigSetString
(
struct ConfigStruct *cfptr,
char *pptr,
char *String,
int SizeOfString
)
{
   register char  *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigSetString()\n");

   zptr = (sptr = String) + SizeOfString;
   while (*pptr && ISLWS(*pptr)) pptr++;
   while (*pptr && sptr < zptr) *sptr++ = *pptr++;
   if (sptr >= zptr)
   {
      *String = '\0';
      ConfigReportProblem (cfptr, NULL, 0);
      return;
   }
   *sptr = '\0';
   if (sptr > String)
   {
      sptr--;
      while (*sptr && ISLWS(*sptr) && sptr > String) sptr--;
      *++sptr = '\0';
   }
}

/*****************************************************************************/
/*
This function formats an error report.  All lines are concatenated onto a
single string of dynamically allocated memory that (obviously) grows as
reports are added to it.  This string is then output if loading the server
configuration or is available for inclusion in an HTML page.
*/

ConfigReportProblem
(
struct ConfigStruct *cfptr,
char *Explanation,
int StatusValue
)
{
   static $DESCRIPTOR (DirectiveFaoDsc,
                       "%HTTPD-W-CONFIG, \"!AZ\" directive !AZ at line !UL\n");
   static $DESCRIPTOR (ExplanationFaoDsc,
                       "%HTTPD-W-CONFIG, \"!AZ\" !AZ at line !UL\n");

   int  status;
   unsigned short  Length;
   char  Buffer [256],
         HtmlBuffer [512];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigReportProblem()\n");

   cfptr->ProblemCount++;

   if (StatusValue)
   {
      if (VMSok (status = sys$getmsg (StatusValue, &Length, &BufferDsc, 0, 0)))
         Buffer[Length++] = '\n';
   }
   if (Explanation == NULL)
      status = sys$fao (&DirectiveFaoDsc, &Length, &BufferDsc,
                        cfptr->DirectiveName, "problem", cfptr->LineNumber);
   else
   if (Explanation[0] == '?')
      status = sys$fao (&DirectiveFaoDsc, &Length, &BufferDsc,
                        cfptr->DirectiveName, "unknown", cfptr->LineNumber);
   else
      status = sys$fao (&ExplanationFaoDsc, &Length, &BufferDsc,
                        cfptr->DirectiveName, Explanation, cfptr->LineNumber);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
      ErrorExitVmsStatus (status, "sys$fao()", FI_LI);
   else
      Buffer[Length] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", Buffer);

   if (cfptr->RequestPtr == NULL) fputs (Buffer, stdout);

   if ((Length = CopyToHtml (HtmlBuffer, sizeof(HtmlBuffer), Buffer, -1)) < 0)
      ErrorExitVmsStatus (SS$_BUFFEROVF, "CopyToHtml()", FI_LI);

   cfptr->ProblemReportPtr =
      VmRealloc (cfptr->ProblemReportPtr, cfptr->ProblemReportLength+Length+1);

   /* include the terminating null */
   memcpy (cfptr->ProblemReportPtr+cfptr->ProblemReportLength,
           HtmlBuffer, Length+1);
   cfptr->ProblemReportLength += Length;
}

/*****************************************************************************/
/*
*/

ConfigSetCommaList
(
struct ConfigStruct *cfptr,
char *pptr,
char **ListPtrPtr,
int *ListLengthPtr
)
{
   register char  *cptr, *sptr;

   int  Comma,
        NewLength,
        ListLength;
   char  *ListPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigSetCommaList()\n");

   if (!*pptr)
   {
      ConfigReportProblem (cfptr, NULL, 0);
      return;
   }

   ListPtr = *ListPtrPtr;
   if (ListLength = *ListLengthPtr) ListLength--;
   if (ListPtr == NULL) Comma = 0; else Comma = 1;
   for (sptr = pptr; *sptr; sptr++);
   NewLength = ListLength + (sptr - pptr) + Comma + 1;
   ListPtr = VmRealloc (ListPtr, NewLength);
   sptr = ListPtr + ListLength;
   if (Comma) *sptr++ = ',';
   while (*pptr) if (ISLWS(*pptr)) pptr++; else *sptr++ = *pptr++;
   *sptr = '\0';
   *ListPtrPtr = ListPtr;
   *ListLengthPtr = NewLength;
   if (Debug)
      fprintf (stdout, "ListPtr %d |%s|\n", *ListLengthPtr, *ListPtrPtr);
}

/*****************************************************************************/
/*
Add a mime content type corresponding to the file suffix (extension, type). 
The content-type can be quoted and can include additional characteristics.

Format: "AddType <suffix> <type[;charset=]> <-|/script> [<description>]"

Examples: "AddType .html text/html - HTML"
          "AddType .html text/html;charset=ISO8859-5 - HTML"
          "AddType .html "text/html; charset=ISO8859-5" - HTML"
          "AddType .txt text/plain - plain text"
          "AddType .hlb text/x-script /conan VMS help library"
          "AddType .gif image/gif - GIF image"

The directive is stored as a linked list of structures of dynamically
allocated memory containing a reformatted version of the directive string.
Function ConfigContentType() searches this list to get pointers to the content
type, any auto-script, and description, into the client request data
structure. A hash table is used to more rapidy index into the list.
*/ 

ConfigAddType
(
struct ConfigStruct *cfptr,
char *pptr
)
{
   register char  *cptr, *sptr;
   register unsigned long  HashValue;
   register struct ContentTypeStruct  *clptr, *hclptr;

   int  status,
        Count = 0;
   char  *DescriptionPtr = "",
         *ContentTypePtr = "",
         *AutoScriptNamePtr = "",
         *SuffixPtr = "";
   char  HtmlDescription [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigAddType() |%s|\n", pptr);

   /* suffix (file type/extension) */
   HashValue = 1;
   SuffixPtr = pptr;
   /* ensure suffix is in upper-case so we don't have to do it matching */
   for (/* above */; *pptr && !ISLWS(*pptr); pptr++) *pptr = toupper(*pptr);
   Count += pptr - SuffixPtr;
   if (*pptr) *pptr++ = '\0';

   /* content type (can include quotes, e.g. "text/html; charset=ISO-8859-5" */
   while (*pptr && ISLWS(*pptr)) pptr++;
   if (*pptr == '\"')
   {
      pptr++;
      ContentTypePtr = pptr;
      while (*pptr && *pptr != '\"') pptr++;
   }
   else
   {
      ContentTypePtr = pptr;
      while (*pptr && !ISLWS(*pptr)) pptr++;
   }
   Count += pptr - ContentTypePtr;
   if (*pptr) *pptr++ = '\0';

   /* (optional) script activation */
   while (*pptr && ISLWS(*pptr)) pptr++;
   AutoScriptNamePtr = pptr;
   while (*pptr && !ISLWS(*pptr)) pptr++;
   Count += pptr - AutoScriptNamePtr;
   if (*pptr) *pptr++ = '\0';

   /* (optional) description, to end of line */
   while (*pptr && ISLWS(*pptr)) pptr++;
   DescriptionPtr = pptr;
   /* find the end-of-line */
   while (*pptr) pptr++;
   if (*pptr) *pptr++ = '\0';
   /* escape any HTML-forbidden characters */
   Count += CopyToHtml (HtmlDescription, sizeof(HtmlDescription),
                        DescriptionPtr, -1);

   if ((*SuffixPtr != '.' && *SuffixPtr != '*') ||
       !*ContentTypePtr || strchr(ContentTypePtr, '/') == NULL ||
       (*AutoScriptNamePtr != '-' &&
        (AutoScriptNamePtr[0] != '/' || !isalpha(AutoScriptNamePtr[1]))))
   {
      ConfigReportProblem (cfptr, NULL, 0);
      return;
   }

   /* plus structure overhead, plus 4 x null terminators */
   Count += cfptr->ContentTypeStructOverhead + 4;
   clptr = (struct ContentTypeStruct*)VmGet (Count);

   sptr = (char*)clptr + cfptr->ContentTypeStructOverhead;

   clptr->SuffixPtr = sptr;
   cptr = SuffixPtr;
   while (*cptr) *sptr++ = *cptr++;
   *sptr++ = '\0';

   clptr->ContentTypePtr = sptr;
   cptr = ContentTypePtr;
   while (*cptr) *sptr++ = *cptr++;
   *sptr++ = '\0';

   clptr->AutoScriptNamePtr = sptr;
   cptr = AutoScriptNamePtr;
   while (*cptr) *sptr++ = *cptr++;
   *sptr++ = '\0';

   clptr->ContentDescriptionPtr = sptr;
   cptr = HtmlDescription;
   while (*cptr) *sptr++ = *cptr++;
   *sptr = '\0';

   /* add to list */
   if (cfptr->ContentTypeListHeadPtr == NULL)
      cfptr->ContentTypeListHeadPtr = clptr; 
   else
      cfptr->ContentTypeListTailPtr->NextPtr = clptr; 
   cfptr->ContentTypeListTailPtr = clptr;
   clptr->NextPtr = NULL;

   /* generates an upper-case hash */
   for (cptr = SuffixPtr; *cptr && *cptr != ';'; cptr++)
      HashValue = (((*cptr)*541)^(HashValue*3)) &
                  CONTENT_TYPE_HASH_TABLE_MASK;
   if (Debug)
      fprintf (stdout, "HashValue: %d %d\n",
               HashValue, cfptr->ContentTypeHashTable[HashValue]);
   if (cfptr->ContentTypeHashTable[HashValue])
   {
      /* add to the hash-collision list */
      for (hclptr = cfptr->ContentTypeHashTable[HashValue];
           hclptr->HashCollisionPtr != NULL;
           hclptr = hclptr->HashCollisionPtr);
      hclptr->HashCollisionPtr = clptr;
   }
   else
      cfptr->ContentTypeHashTable[HashValue] = clptr;

   /* special case, default content-type for unknown suffixes */
   if (*(unsigned short*)clptr->SuffixPtr == '*\0')
   {
      cfptr->ContentTypeDefaultPtr = VmGet (strlen(DescriptionPtr+1));
      strcpy (cfptr->ContentTypeDefaultPtr, DescriptionPtr);
      clptr->ContentTypeUnknown = true;
      /* cancel type description */
      *clptr->ContentDescriptionPtr = '\0';
   }
   else
      clptr->ContentTypeUnknown = false;
}

/*****************************************************************************/
/*
After loading all content-types (i.e. after configuration) this function sets
the corresponding icon for each type.
*/ 

ConfigContentTypeIcon (struct ConfigStruct *cfptr)

{
   register struct ContentTypeStruct  *clptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigContentTypeIcon()\n");

   for (clptr = cfptr->ContentTypeListHeadPtr;
        clptr != NULL;
        clptr = clptr->NextPtr)
   {
      if (Debug) fprintf (stdout, "clptr: %d\n", clptr);
      if (clptr->ContentTypeUnknown)
         clptr->IconPtr = ConfigIconFor (ConfigContentTypeUnknown);
      else
         clptr->IconPtr = ConfigIconFor (clptr->ContentTypePtr);
   }

   ConfigBlankIconPtr = ConfigIconFor (ConfigContentTypeBlank);
   ConfigDirIconPtr = ConfigIconFor (ConfigContentTypeDir);
   ConfigParentIconPtr = ConfigIconFor (ConfigContentTypeParent);
   ConfigUnknownIconPtr = ConfigIconFor (ConfigContentTypeUnknown);
}

/*****************************************************************************/
/*
Using the hash table this function searches the list of file suffixes (types,
extensions) and content-types, returning a pointer to the content-type string
and if supplied with a content-type structure fills out the appropriate
fields.
*/ 

char* ConfigContentType
(
struct ContentTypeStruct *ctptr,
char *Suffix
)
{
   register char  *cptr, *sptr;
   register unsigned long  HashValue;
   register struct ContentTypeStruct  *clptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigContentType() |%s|\n", Suffix);

   /* generate an upper-case hash */
   HashValue = 1;
   for (cptr = Suffix; *cptr && *cptr != ';'; cptr++)
      HashValue = ((toupper(*cptr)*541)^(HashValue*3)) &
                  CONTENT_TYPE_HASH_TABLE_MASK;
   if (Debug)
      fprintf (stdout, "HashValue: %d %d\n",
               HashValue, Config.ContentTypeHashTable[HashValue]);

   clptr = Config.ContentTypeHashTable[HashValue];
   while (clptr != NULL)
   {
      if (Debug) fprintf (stdout, "clptr: %d\n", clptr);

      /* 'clptr->SuffixPtr' has already been converted to upper case! */
      cptr = clptr->SuffixPtr;
      if (Debug) fprintf (stdout, "|%s|\n", cptr);
      sptr = Suffix;
      while (*cptr && *sptr && *sptr != ';' && *cptr == toupper(*sptr))
         { cptr++; sptr++; }

      if (*cptr || (*sptr && *sptr != ';'))
      {
         /* didn't match, traversing the hash-collision list */
         clptr = clptr->HashCollisionPtr;
         continue;
      }

      /**********/
      /* match! */
      /**********/

      /* if just getting a pointer to a content type then just return that */
      if (ctptr == NULL) return (clptr->ContentTypePtr);

      ctptr->ContentTypeUnknown = false;
      ctptr->IconPtr = clptr->IconPtr;
      ctptr->ContentTypePtr = clptr->ContentTypePtr;
      ctptr->AutoScriptNamePtr = clptr->AutoScriptNamePtr;
      ctptr->ContentDescriptionPtr = clptr->ContentDescriptionPtr;

      if (Debug)
         fprintf (stdout, "|%s|%s|%s|\n",
                  ctptr->ContentTypePtr, ctptr->AutoScriptNamePtr,
                  ctptr->ContentDescriptionPtr);

      return (clptr->ContentTypePtr);
   }

   /*************/
   /* no match! */
   /*************/

   /* if just getting a pointer to a content type then just return that */
   if (ctptr == NULL) return (ConfigContentTypeUnknown);

   ctptr->ContentTypeUnknown = true;
   ctptr->IconPtr = ConfigUnknownIconPtr;
   ctptr->ContentTypePtr = ConfigContentTypeUnknown;
   ctptr->AutoScriptNamePtr = ctptr->ContentDescriptionPtr = "";
   return (ConfigContentTypeUnknown);
}

/****************************************************************************/
/*
Compare two MIME content-types where only the type is compared (i.e. only the
"text/html" from something like "text/html;charset=ISO8859-5").  It works much
the same as strsame(), except a semi-colon or white-space in either string will
terminate a comparison, as will an exhausted count value.
*/ 
 
boolean ConfigSameContentType
(
register char *sptr1,
register char *sptr2,
register int  count
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "ConfigSameContentType() |%s|%s| %d\n",
               sptr1, sptr2, count);

   while (*sptr1 && *sptr1 != ';' && !ISLWS(*sptr1) &&
          *sptr2 && *sptr2 != ';' && !ISLWS(*sptr2))
   {
      if (toupper (*sptr1++) != toupper (*sptr2++)) return (false);
      if (count)
         if (!--count) return (true);
   }
   if ((!*sptr1 && (!*sptr2 || *sptr2 == ';')) ||
       (!*sptr2 && (!*sptr1 || *sptr1 == ';')))
      return (true);
   else
      return (false);
}
 
/*****************************************************************************/
/*
Add a URL for an icon corresponding to the mime content type.

Format:   "AddIcon <URL> <alternative-text> <content-type-template>"

Examples: "AddIcon /icon/httpd/text.xbm [TXT] text/plain"
          "AddIcon /icon/httpd/doc.xbm [HTM] text/html"
          "AddIcon /icon/httpd/image.xbm [IMG] image/jpeg"

The template can contain a wildcard in place of a string in either side of the 
template slash.  In this case the wildcard matches any string on that side.

The directive is stored as a linked-list of structures of dynamically
allocated memory containing a reformatted version of the directive string,
including a complete HTML anchor for the URL. Function ConfigIconFor()
searches this list to return a pointer to the anchor. A hash table is used to
more rapidy index into the list.
*/ 

ConfigAddIcon
(
struct ConfigStruct *cfptr,
char *pptr
)
{
   register unsigned long  HashValue;
   register char  *cptr;
   register struct IconStruct  *ilptr, *hilptr;

   int  Count = 0;
   char  *AltTextPtr,
         *IconUrlPtr,
         *ContentTypePtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigAddIcon()\n");

   IconUrlPtr = pptr;
   while (*pptr && !ISLWS(*pptr)) pptr++;
   Count = pptr - IconUrlPtr;
   if (*pptr) *pptr++ = '\0';

   /* alternative text */
   while (*pptr && ISLWS(*pptr)) pptr++;
   AltTextPtr = pptr;
   while (*pptr && !ISLWS(*pptr))
   {
      /* underscores represent spaces! (especially for the "blank" icon) */
      if (*pptr == '_') *pptr = ' ';
      pptr++;
   }
   Count += pptr - AltTextPtr;
   if (*pptr) *pptr++ = '\0';

   /* template */
   while (*pptr && ISLWS(*pptr)) pptr++;
   ContentTypePtr = pptr;
   /* force to lower case to save a tolower() in ConfigIconFor() */
   for (/* above */; *pptr && !ISLWS(*pptr); pptr++)
      *pptr = tolower(*pptr);
   Count += pptr - ContentTypePtr;
   if (*pptr) *pptr++ = '\0';

   if (!*IconUrlPtr && *IconUrlPtr != '/' ||
       !*AltTextPtr ||
       !*ContentTypePtr || strchr(ContentTypePtr, '/') == NULL)
   {
      ConfigReportProblem (cfptr, NULL, 0);
      return;
   }

   /* plus structure overhead, plus the HTML IMG element used below */
   Count += cfptr->IconStructOverhead + 32;
   ilptr = (struct IconStruct*)VmGet (Count);

   sprintf ((char*)ilptr + cfptr->IconStructOverhead,
            "%s <IMG ALIGN=top SRC=\"%s\" ALT=\"%s\">",
            ContentTypePtr, IconUrlPtr, AltTextPtr);
   if (Debug)
      fprintf (stdout, "|%s|%s|%s|%s|\n",
               ContentTypePtr, IconUrlPtr, AltTextPtr,
               (char*)ilptr + cfptr->IconStructOverhead);

   /* add to list */
   if (cfptr->IconListHeadPtr == NULL)
      cfptr->IconListHeadPtr = ilptr; 
   else
      cfptr->IconListTailPtr->NextPtr = ilptr; 
   cfptr->IconListTailPtr = ilptr;
   ilptr->NextPtr = NULL;

   /* generates a lower-case hash */
   HashValue = 1;
   for (cptr = ContentTypePtr; *cptr; cptr++)
      HashValue = (((*cptr)*541)^(HashValue*3)) &
                  ICON_HASH_TABLE_MASK;
   if (Debug)
      fprintf (stdout, "HashValue: %d %d\n",
               HashValue, cfptr->ContentTypeHashTable[HashValue]);
   if (cfptr->IconHashTable[HashValue])
   {
      /* add to the hash-collision list */
      for (hilptr = cfptr->IconHashTable[HashValue];
           hilptr->HashCollisionPtr != NULL;
           hilptr = hilptr->HashCollisionPtr);
      hilptr->HashCollisionPtr = ilptr;
   }
   else
      cfptr->IconHashTable[HashValue] = ilptr;
}

/*****************************************************************************/
/*
Using the hash table this function searches the icons for one corresponding to
the supplied mime content type. The directive content type and associated HTML
anchor (containing the original configuration-supplied URL) is stored as an
list of structures dynamically allocated memory containing a string.
*/

char* ConfigIconFor (char *ContentType)

{
   static char  IconNotFound [] = "<FONT COLOR=\"#ff0000\"><B>[?]</B></FONT>";

   register char  *cptr, *sptr;
   register unsigned long  HashValue;
   register struct IconStruct  *ilptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigIconFor() |%s|\n", ContentType);

   /* generates a lower-case hash */
   HashValue = 1;
   for (cptr = ContentType; *cptr && *cptr != ';' && !ISLWS(*cptr); cptr++)
      HashValue = ((tolower(*cptr)*541)^(HashValue*3)) &
                  ICON_HASH_TABLE_MASK;
   if (Debug)
      fprintf (stdout, "HashValue: %d %d\n",
               HashValue, Config.IconHashTable[HashValue]);

   ilptr = Config.IconHashTable[HashValue];
   while (ilptr != NULL)
   {
      if (Debug) fprintf (stdout, "ilptr: %d\n", ilptr);

      cptr = (char*)ilptr + Config.IconStructOverhead;
      if (Debug) fprintf (stdout, "|%s|\n", cptr);
      sptr = ContentType;
      if (*cptr == '*')
      {
         /* template contained a wildcard on the left side of the slash */
         while (*cptr && *cptr != '/' && !ISLWS(*cptr)) cptr++;
         if (*cptr == '/') cptr++;
         if (*cptr == '*')
         {
            /* return a pointer to the HTML IMG element for the icon */
            cptr++;
            return (cptr);
         }
         /* scan to subtype */
         while (*sptr && *sptr != '/' && *sptr != ';' && !ISLWS(*sptr)) sptr++;
      }
      else
      {
         /* match main type */
         while (*cptr == tolower(*sptr) &&
                *cptr && *cptr != '/' &&
                *sptr && *sptr != '/' && *sptr != ';' && !ISLWS(*sptr))
         {
            cptr++;
            sptr++;
         }
         if (*cptr != '/' && *sptr != '/')
         {
            /* didn't match, traversing the hash-collision list */
            ilptr = ilptr->HashCollisionPtr;
            continue;
         }
         /* step across the '/' to the subtype */
         cptr++;
         sptr++;
      }

      if (*cptr == '*')
      {
         /* wildcard subtype, return pointer to the IMG element for the icon */
         cptr++;
         return (cptr);
      }
      else
      {
         /* match subtype */
         while (*cptr == tolower(*sptr) &&
                *cptr && !ISLWS(*cptr) &&
                *sptr && *sptr != ';' && !ISLWS(*sptr))
         {
            cptr++;
            sptr++;
         }
         if (!ISLWS(*cptr) || (*sptr && *sptr != ';' && !ISLWS(*sptr)))
         {
            /* didn't match, traversing the hash-collision list */
            ilptr = ilptr->HashCollisionPtr;
            continue;
         }
         /* return pointer to the IMG element for the icon */
         cptr++;
         return (cptr);
      }
   }

   return (IconNotFound);
}

/*****************************************************************************/
/*
Accept or reject the client connection.  Lists of hosts must be a comma-
separated string.  A wildcard '*' in a domain specification matches any string 
that occurs between the dots.
Example: "host1.network1.domain1,*.network2.domain1,*.*.domain2". 
*/ 

boolean ConfigAcceptClientHostName
(
char *ClientIpAddressString,
char *ClientHostName
)
{
   register char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "ConfigAcceptConnection() |%s|\n", ClientHostName);

   if (Config.AcceptHostsPtr == NULL && Config.RejectHostsPtr == NULL)
      return (true);

   if (!ClientHostName[0] || ClientHostName[0] == '?') return (false);

   /******************/
   /* hosts accepted */
   /******************/

   if (Config.AcceptHostsPtr != NULL)
   {
      sptr = Config.AcceptHostsPtr;
      while (*sptr)
      {
         /** if (Debug) fprintf (stdout, "sptr |%s|\n", sptr); **/
         if (isdigit(*sptr))
            cptr = ClientIpAddressString;
         else
            cptr = ClientHostName;
         while (*cptr)
         {
            if (*sptr == ',') break;
            if (*sptr == '*')
            {
               while (*sptr && *sptr == '*' && *sptr != ',') sptr++;
               while (*cptr && *cptr != *sptr) cptr++;
            }
            if (tolower(*cptr) != tolower(*sptr)) break;
            if (*cptr) cptr++;
            if (*sptr) sptr++;
         }
         if (!*cptr && (!*sptr || *sptr == ',')) break;
         while (*sptr && *sptr != ',') sptr++;
         if (*sptr) sptr++;
      }
      /* if it was NOT found then reject the connection request */
      if (!(!*cptr && (!*sptr || *sptr == ',')))
         return (false);
   }

   /******************/
   /* hosts rejected */
   /******************/

   if (Config.RejectHostsPtr != NULL)
   {
      sptr = Config.RejectHostsPtr;
      while (*sptr)
      {
         /** if (Debug) fprintf (stdout, "sptr |%s|\n", sptr); **/
         cptr = ClientHostName;
         while (*cptr)
         {
            if (*sptr == ',') break;
            if (*sptr == '*')
            {
               while (*sptr && *sptr == '*' && *sptr != ',') sptr++;
               while (*cptr && *cptr != *sptr) cptr++;
            }
            if (tolower(*cptr) != tolower(*sptr)) break;
            if (*cptr) cptr++;
            if (*sptr) sptr++;
         }
         if (!*cptr && (!*sptr || *sptr == ',')) break;
         while (*sptr && *sptr != ',') sptr++;
         if (*sptr) sptr++;
      }
      /* if it WAS found then reject the connection request */
      if (!*cptr && (!*sptr || *sptr == ','))
         return (false);
   }

   /* accept the connection */
   return (true);
}

/*****************************************************************************/
/*
Set directory access booleans.  Allowed values OFF, ON, SELECTIVE.
*/

ConfigSetDirAccess
(
struct ConfigStruct *cfptr,
char *pptr
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigSetDirAccess()\n");

   if (strsame (pptr, "YES", 3) ||
       strsame (pptr, "ON", 2) ||
       strsame (pptr, "ENABLED", 7))
   {
      cfptr->DirAccess = true;
      cfptr->DirAccessSelective = false;
   }
   else
   if (strsame (pptr, "NO", 2) ||
       strsame (pptr, "OFF", 3) ||
       strsame (pptr, "DISABLED", 8))
      cfptr->DirAccess = cfptr->DirAccessSelective = false;
   else
   if (strsame (pptr, "SELECTIVE", 9))
      cfptr->DirAccess = cfptr->DirAccessSelective = true;
   else
      ConfigReportProblem (cfptr, NULL, 0);
}

/*****************************************************************************/
/*
Set directory README file booleans.  Allowed values OFF, TOP, BOTTOM.
*/

ConfigSetDirReadMe
(
struct ConfigStruct *cfptr,
char *pptr
)
{
   if (Debug) fprintf (stdout, "ConfigSetDirReadMe()\n");

   if (strsame (pptr, "NO", 2) ||
       strsame (pptr, "OFF", 3) ||
       strsame (pptr, "DISABLED", 8))
      cfptr->DirReadMeBottom = cfptr->DirReadMeTop = false;
   else
   if (strsame (pptr, "BOTTOM", 6))
   {
      cfptr->DirReadMeTop = false;
      cfptr->DirReadMeBottom = true;
   }
   else
   if (strsame (pptr, "TOP", 3))
   {
      cfptr->DirReadMeTop = true;
      cfptr->DirReadMeBottom = false;
   }
   else
      ConfigReportProblem (cfptr, NULL, 0);
}

/*****************************************************************************/
/*
Add a "read-me" file name.

Format:   "ReadMeFile file.suffix"

Examples: "ReadMeFile README.HTML"
          "ReadMeFile README.TXT"
          "ReadMeFile README."

The directive is stored as an array of pointers to dynamically allocated 
memory containing the "read-me" file name.  Function ConfigReadMeFile() gets 
these by index number.
*/

ConfigSetDirReadMeFile
(
struct ConfigStruct *cfptr,
char *pptr
)
{
   int  status,
        Count;
   char  *NamePtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigSetDirReadMeFile()\n");

   if (cfptr->ReadMeFileCount >= CONFIG_README_FILES_MAX)
   {
      ConfigReportProblem (cfptr, "maximum read-me files", 0);
      return;
   }

   NamePtr = pptr;
   while (*pptr && !ISLWS(*pptr))
   {
      *pptr = toupper(*pptr);
      pptr++;
   }
   Count = pptr - NamePtr;
   *pptr = '\0';

   if (!*NamePtr)
   {
      ConfigReportProblem (cfptr, NULL, 0);
      return;
   }

   if (Debug) fprintf (stdout, "VmGet()ing %d bytes\n", Count+1);
   cfptr->ReadMeFileArray[cfptr->ReadMeFileCount] = VmGet(Count+1);

   strcpy (cfptr->ReadMeFileArray[cfptr->ReadMeFileCount], NamePtr);
   if (Debug)
      fprintf (stdout, "|%s|\n",
               cfptr->ReadMeFileArray[cfptr->ReadMeFileCount]);
   cfptr->ReadMeFileCount++;
}

/*****************************************************************************/
/*
Return a pointer to the "read-me" file name stored against the index number of 
the array.  If none exists return a null string.  The calling function would 
use this by stepping through the index numbers 0 to n, until the null string 
was returned indicating the possible "read-me" file names were exhausted.
*/

char* ConfigReadMeFile (int Number)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigReadMeFile |%d|\n", Number);

   if (Number >= 0 && Number < Config.ReadMeFileCount)
      return (Config.ReadMeFileArray[Number]);
   else
      return ("");
}

/*****************************************************************************/
/*
Add a script type.

Format:   "file-type verb"

Examples: ".PL PERL"
          ".PL $PERL_EXE:PERL.EXE"

The directive is stored as an array of pointers to dynamically allocated 
memory containing definition.  Function DclFindScript() uses this information.
The array entries comprise a reconstructed string from the directive.  The
verb part of the directive can have a leading '@' or '$' indicating that the
verb should be create before invoking the script.  If not having these leading
characters the verb is considered to be globally available.  The leading '@'
and '$' verbs are reconstructed from "$perl_exe:perl.exe" into
"PERL=\"$perl_exe:perl.exe\"".  From this the DCL module creates an
excecutable DCL command sequence.
*/ 

ConfigSetScriptRunTime
(
struct ConfigStruct *cfptr,
char *pptr
)
{
   register char  *cptr, *sptr;

   int  status,
        Count;
   char  Scratch [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigSetScriptRunTime()\n");

   if (cfptr->ScriptRunTimeCount >= CONFIG_SCRIPT_RUNTIME_MAX)
   {
      ConfigReportProblem (cfptr, "maximum script types", 0);
      return;
   }

   sptr = Scratch;
   if (*pptr != '.') *sptr++ = '.';
   while (*pptr && !ISLWS(*pptr)) *sptr++ = toupper(*pptr++);
   if (sptr > Scratch && *(sptr-1) != ';') *sptr++ = ';';
   cfptr->ScriptRunTimeLengthArray[cfptr->ScriptRunTimeCount] = sptr - Scratch;
   *sptr++ = ' ';
   while (*pptr && ISLWS(*pptr)) pptr++;
   if (!*pptr)
   {
      ConfigReportProblem (cfptr, NULL, 0);
      return;
   }
   if (*pptr == '@' || *pptr == '$')
   {
      *sptr++ = *(cptr = pptr);
      while (*cptr != ':') cptr++;
      if (*cptr) cptr++;
      if (*cptr == '[')
      {
         while (*cptr && *cptr != ']') cptr++;
         if (*cptr) cptr++;
      }
      while (*cptr && *cptr != '.' && !ISLWS(*cptr))
         *sptr++ = toupper(*cptr++);
      *sptr++ = '=';
      *sptr++ = '\"';
      while (*pptr)
      {
         /* escape quotes for DCL */
         if (*pptr == '\"') *sptr++ = '\"';
         *sptr++ = *pptr++;
      }
      *sptr++ = '\"';
      *sptr++ = '\0';
      Count = sptr - Scratch;
   }
   else
   {
      while (*pptr) *sptr++ = *pptr++;
      *sptr++ = '\0';
      Count = sptr - Scratch;
   }

   if (Debug) fprintf (stdout, "VmGet()ing %d bytes\n", Count);
   cfptr->ScriptRunTimeArray[cfptr->ScriptRunTimeCount] = VmGet (Count);

   strcpy (cfptr->ScriptRunTimeArray[cfptr->ScriptRunTimeCount], Scratch);
   if (Debug)
      fprintf (stdout, "|%s|\n",
               cfptr->ScriptRunTimeArray[cfptr->ScriptRunTimeCount]);
   cfptr->ScriptRunTimeCount++;
}

/*****************************************************************************/
/*
Add a welcome (home) page file name.

Format:   "Welcome file.suffix"

Examples: "Welcome HOME.HTML"
          "Welcome HOME.HTM"
          "Welcome HOME.MENU"

The directive is stored as an array of pointers to dynamically allocated 
memory containing the home page file name.  Function ConfigHomePage() gets 
these by index number.
*/ 

ConfigSetWelcome
(
struct ConfigStruct *cfptr,
char *pptr
)
{
   int  status,
        Count;
   char  *WelcomePtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigSetWelcome()\n");

   if (cfptr->HomePageCount >= CONFIG_HOME_PAGES_MAX)
   {
      ConfigReportProblem (cfptr, "maximum welcome files", 0);
      return;
   }

   WelcomePtr = pptr;
   while (*pptr && !ISLWS(*pptr))
   {
      *pptr = toupper(*pptr);
      pptr++;
   }
   Count = pptr - WelcomePtr;
   *pptr = '\0';

   if (!*WelcomePtr)
   {
      ConfigReportProblem (cfptr, NULL, 0);
      return;
   }

   if (Debug) fprintf (stdout, "VmGet()ing %d bytes\n", Count+1);
   cfptr->HomePageArray[cfptr->HomePageCount] = VmGet (Count+1);

   strcpy (cfptr->HomePageArray[cfptr->HomePageCount], WelcomePtr);
   if (Debug)
      fprintf (stdout, "|%s|\n", cfptr->HomePageArray[cfptr->HomePageCount]);
   cfptr->HomePageCount++;
}

/*****************************************************************************/
/*
Return a pointer to the home page (welcome) file name stored against the index 
number of the array.  If none exists return a null string.  The calling 
function would use this by stepping through the index numbers 0 to n, until 
the null string was returned indicating the possible home page file names were 
exhausted.
*/

char* ConfigHomePage (int Number)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigHomePage |%d|\n", Number);

   if (Number >= 0 && Number < Config.HomePageCount)
      return (Config.HomePageArray[Number]);
   else
      return ("");
}

/*****************************************************************************/
/*
A server administration report on the server's configuration. This function
just wraps the reporting function, loading a temporary database if necessary
for reporting from the configuration file.
*/ 

ConfigReport
(
struct RequestStruct *rqptr,
void *NextTaskFunction,
boolean  UseServerDatabase
)
{
   int  status;
   struct ConfigStruct  LocalConfig;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigReport()\n");

   if (UseServerDatabase)
      ConfigReportNow (rqptr, &Config);
   else
   {
      memset (&LocalConfig, 0, sizeof(struct ConfigStruct));
      /* indicate it's being used for a report */
      Config.RequestPtr = rqptr;
      Configure (&LocalConfig);
      ConfigReportNow (rqptr, &LocalConfig);
      ConfigFree (&LocalConfig);
   }

   NetWriteBuffered (rqptr, NextTaskFunction, NULL, 0);
}

/*****************************************************************************/
/*
Return a report on the HTTPd server's configuration ... now!  This function
blocks while executing.
*/ 

ConfigReportNow
(
struct RequestStruct *rqptr,
struct ConfigStruct *cfptr
)
{
/* macro to push booleans onto the sys$faol() parameter list */
#define REPBOOL(b) \
if (b) *vecptr++ = "[enabled]"; else *vecptr++ = "[disabled]";

   static $DESCRIPTOR (ResponseFaoDsc,
"<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>HTTPd !AZ ... Server Configuration</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1><NOBR>HTTPd !AZ</NOBR></H1>\n\
<H2>Server Configuration</H2>\n\
!AZ\n");

   static $DESCRIPTOR (ProblemReportFaoDsc,
"<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH><FONT COLOR=\"#ff0000\">!UL Report!AZ !AZ</FONT></TH></TR>\n\
<TR><TD><PRE>!AZ</PRE></TD></TR>\n\
</TABLE>\n");

   static $DESCRIPTOR (ReportSourceFaoDsc,
"<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Source: &quot;!AZ&quot;</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right>File:</TH>\
<TD ALIGN=left>!AZ</TD>\
<TD>&nbsp;</TD><TH>[<A HREF=\"!AZ\">View</A>]</TH>\
<TH>[<A HREF=\"!AZ!AZ\">Edit</A>]</TH>\
</TR>\n\
<TR><TH ALIGN=right>!AZ</TH>\
<TD ALIGN=left>!AZ</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n");

   static char
      NotNone [] = "<FONT SIZE=-1><I>(none)</I></FONT>\n",
      NoRestriction [] = "<FONT SIZE=-1><I>(no restriction)</I></FONT>\n";

   static $DESCRIPTOR (ReportGeneralFaoDsc,
"<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>GMT</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right>Difference:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Current:</TH><TD>!AZ</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>General</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right>Busy:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>connections</I></FONT></TR>\
<TR><TH ALIGN=right>Request History:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>requests</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>Activity History:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>days</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>Charset Default:</TH><TD COLSPAN=2>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Comments:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>DNS Lookup:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Error Report Path:</TH><TD COLSPAN=2>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Error Recommend:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Error Source Information:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Monitor:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>IP Port:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right>StreamLF:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>kBytes</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>Search Script:</TH><TD COLSPAN=2>!AZ</TD></TR>\n\
<TR><TH ALIGN=right><FONT SIZE=-1>PUT/POST</FONT> Max Length:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>kBytes</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right><FONT SIZE=-1>PUT/POST</FONT> Files:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>version limit</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>/~username/ Cache Entries:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>user names</I></FONT></TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Log</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right>Logging:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Format:</TH><TD>!AZ</TD>\
<TR><TH ALIGN=right>Naming:</TH><TD>!AZ</TD>\
<TR><TH ALIGN=right>Period:</TH><TD>!AZ</TD>\
<TR><TH ALIGN=right>Per-Service:</TH><TD>!AZ</TD>\
<TR><TH ALIGN=right>File:</TH><TD>!AZ</TD>\
<TR><TH ALIGN=right>Exclude Hosts:</TH><TD>!AZ</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>File Cache</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right>Caching:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Hash Table</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>entries</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>Max Entries:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>files</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>Max Cache Size:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>kBytes</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>Max File Size:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>kBytes</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>Chunk Size:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>kBytes</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>Validate:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>seconds</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>Frequent:</TH><TD COLSPAN=2>\
!UL &nbsp;<FONT SIZE=-1><I>hits, last within</I></FONT>&nbsp; \
!UL &nbsp;<FONT SIZE=-1><I>seconds</I></FONT></TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Timeouts</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right>Input:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>minutes</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>Output:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>minutes</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>No-Progress:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>minutes</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>Keep-Alive:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>seconds</I></FONT></TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Buffer Sizes</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right>DCL <FONT SIZE=-1>SYS$COMMAND</FONT>:</TH>\
<TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right>DCL <FONT SIZE=-1>SYS$OUTPUT</FONT>:</TH>\
<TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right>DCL <FONT SIZE=-1>CGIPLUSIN</FONT>:</TH>\
<TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right>Network Read:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right>Network Write:</TH><TD>!UL</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Welcome (Home) Pages</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n");

   static $DESCRIPTOR (ReportHostsFaoDsc,
"</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Hosts</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right>Accept:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Reject:</TH><TD>!AZ</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Proxy Serving</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right>Proxy Serving:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Cache:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Add &quot;Forwarded: by&quot;:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Report Events To Log:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Report Cache Events To Log:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Host Name Lookup Retry:</TH><TD>!UL</TD>\
<TR><TH ALIGN=right>Host Name Cache Purge:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>hours</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right><U>Cache</U></TH></TR>\n\
<TR><TH ALIGN=right>Max Load:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>kBytes</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>Reload List:</TH><TD>!AZ</TD>\
<TD><FONT SIZE=-1><I>hours</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>Routine Check:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>hour-of-day</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>Purge List:</TH><TD>!AZ</TD>\
<TD><FONT SIZE=-1><I>hours</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right><U>Cache Device</U></TH></TR>\n\
<TR><TH ALIGN=right>Usage Check:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>minutes</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>Usage Max:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>percent</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>Reduce By:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>percent</I></FONT></TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Authentication</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right>Cache Valid:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>minutes</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>Revalidate User:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>minutes</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>Failure Limit:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>attempts</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>Basic</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Digest:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Digest GET lifetime:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>minutes</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>Digest PUT lifetime:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>minutes</I></FONT></TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Scripting &nbsp;(DCL & DECnet)</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right>Spawn Authorized Privileges:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Supply Header and Body:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Soft Limit:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>subprocesses</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>Hard Limit:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>subprocesses</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>Zombie Life-Time:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>minutes</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>CGIplus Life-Time:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>minutes</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>DECnet Reuse Life-Time:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>minutes</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>DECnet Connect List Max:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>connections</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>Script Run-Time:</TH>\
<TD ALIGN=left><U>file-type</U></TD><TD ALIGN=left><U>verb</U></TD></TR>\n");

   static $DESCRIPTOR (ReportSsiFaoDsc,
"</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
</TABLE>\n\
</TD></TR>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Server-Side Includes</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right>SSI:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Exec:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Accesses:</TH><TD>!AZ</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Directory</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right>Access:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Layout:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>HTML Description:</TH><TD>!UL</TD>\
<TD><FONT SIZE=-1><I>lines</I></FONT></TD></TR>\n\
<TR><TH ALIGN=right>Owner:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Pre-Expired:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Wildcard:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>No Implied Wildcard:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Prot Violations Ignored:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Readme:</TH><TD>!AZ &nbsp;!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Readme Files:</TH>\n");

   static $DESCRIPTOR (ReportIconsFaoDsc,
"</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH COLSPAN=4>Directory Icons</TH></TR>\n\
<TR><TH>content-type</TH><TH>icon</TH><TH>path</TH>\
<TH>alternate text</TH></TR>\n");

   static $DESCRIPTOR (ReportContentTypeFaoDsc,
"</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH COLSPAN=4>Content-Type</TH></TR>\n\
<TR><TH>suffix</TH><TH>content-type</TH><TH>icon</TH>\
<TH>auto-script</TH><TH>description</TH></TR>\n");

   static $DESCRIPTOR (ContentTypeFaoDsc,
"<TR><TD>!AZ</TD><TD>!AZ</TD><TD>!AZ</TD><TD>!AZ</TD><TD>!AZ</TD></TR>\n");

   static char  EndPage [] =
"</TABLE>\n\
\
</BODY>\n\
</HTML>\n";

   register int  idx;
   register unsigned long  *vecptr;
   register char  *cptr, *sptr;
   register struct ContentTypeStruct  *clptr;
   register struct IconStruct  *ilptr;

   int  status,
        ServiceListCount;
   unsigned short  Length;
   unsigned long  FaoVector [128];
   char  Buffer [8192],
         HtmlIconAltText [256],
         IconAltText [256],
         IconPath [256];
   char  *AutoScriptNamePtr,
         *ContentDescriptionPtr;
   struct ContentTypeStruct  ContentType;
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigReportNow()\n");

   rqptr->ResponsePreExpired = PRE_EXPIRE_ADMIN;
   HTTP_HEADER_200_HTML (rqptr);

   vecptr = FaoVector;

   *vecptr++ = HtmlMetaInfo (rqptr, NULL);

   *vecptr++ = ServerHostPort;
   *vecptr++ = ServerHostPort;
   *vecptr++ = DayDateTime (&rqptr->BinaryTime, 20);

   status = sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      return;
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   if (cfptr->ProblemReportLength)
   {
      vecptr = FaoVector;
      if ((*vecptr++ = cfptr->ProblemCount) == 1)
         *vecptr++ = "";
      else
         *vecptr++ = "s";
      if (cfptr == &Config)
         *vecptr++ = "At Startup";
      else
         *vecptr++ = "During Load";
      *vecptr++ = cfptr->ProblemReportPtr;

      status = sys$faol (&ProblemReportFaoDsc, &Length, &BufferDsc,
                         &FaoVector);
      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         rqptr->ErrorTextPtr = "sys$faol()";
         ErrorVmsStatus (rqptr, status, FI_LI);
         return;
      }

      Buffer[Length] = '\0';
      NetWriteBuffered (rqptr, 0, Buffer, Length);
   }

   cptr = MapVmsPath (cfptr->LoadFileName, rqptr);

   vecptr = FaoVector;

   /* source file */
   if (cfptr == &Config)
      *vecptr++ = "Server";
   else
      *vecptr++ = "File";
   *vecptr++ = cfptr->LoadFileName;
   *vecptr++ = cptr;
   *vecptr++ = INTERNAL_SCRIPT_UPD;
   *vecptr++ = cptr;
   if (cfptr == &Config)
   {
      *vecptr++ = "Loaded:";
      *vecptr++ = DayDateTime (&cfptr->LoadBinTime, 20);
   }
   else
   {
      *vecptr++ = "Revised:";
      *vecptr++ = DayDateTime (&cfptr->RevBinTime, 20);
   }

   status = sys$faol (&ReportSourceFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      return;
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   /************/
   /* services */
   /************/

   if (VMSnok (NetServiceReportConfig (rqptr, cfptr)))
      return;

   /***********/
   /* general */
   /***********/

   vecptr = FaoVector;

   /* GMT */
   *vecptr++ = TimeGmtString;
   *vecptr++ = rqptr->GmDateTime;

   /* general */
   *vecptr++ = cfptr->Busy;
   *vecptr++ = cfptr->RequestHistory;
   *vecptr++ = cfptr->ActivityNumberOfDays;
   if (cfptr->CharsetDefault[0])
      *vecptr++ = cfptr->CharsetDefault;
   else
      *vecptr++ = NotNone;
   REPBOOL (cfptr->IncludeCommentedInfo)
   REPBOOL (cfptr->DnsLookup)
   if (cfptr->ErrorReportPath[0])
      *vecptr++ = cfptr->ErrorReportPath;
   else
      *vecptr++ = NotNone;
   REPBOOL (cfptr->ErrorRecommend)
   REPBOOL (cfptr->ErrorSourceInfo)
   REPBOOL (cfptr->MonitorEnabled)
   *vecptr++ = cfptr->ServerPort;
   *vecptr++ = cfptr->StreamLfConversionMaxKbytes;
   *vecptr++ = cfptr->KeywordSearchScript;
   *vecptr++ = cfptr->PutMaxKbytes;
   *vecptr++ = cfptr->PutVersionLimit;
   if (cfptr == &Config)
      *vecptr++ = MapUrlUserNameCacheEntries;
   else
      *vecptr++ = cfptr->MapUserNameCacheEntries;

   /* logging */
   REPBOOL (cfptr->LoggingEnabled)
   if (cfptr->LogFormat[0])
      *vecptr++ = cfptr->LogFormat;
   else
      *vecptr++ = NotNone;
   if (cfptr->LogNaming[0])
      *vecptr++ = cfptr->LogNaming;
   else
      *vecptr++ = NotNone;
   if (cfptr->LogPeriod[0])
      *vecptr++ = cfptr->LogPeriod;
   else
      *vecptr++ = NotNone;
   REPBOOL (cfptr->LogPerService)
   if (cfptr->LogFileName[0])
      *vecptr++ = cfptr->LogFileName;
   else
      *vecptr++ = NotNone;

   if (cfptr->LogExcludeHostsPtr == NULL)
      *vecptr++ = NotNone;
   else
      *vecptr++ = ConfigCommaList (rqptr, cfptr->LogExcludeHostsPtr, ',');

   /* cache */
   REPBOOL (cfptr->CacheEnabled)
   *vecptr++ = cfptr->CacheHashTableEntries;
   *vecptr++ = cfptr->CacheEntriesMax;
   *vecptr++ = cfptr->CacheTotalKBytesMax;
   *vecptr++ = cfptr->CacheFileKBytesMax;
   *vecptr++ = cfptr->CacheChunkKBytes;
   *vecptr++ = cfptr->CacheValidateSeconds;
   *vecptr++ = cfptr->CacheFrequentHits;
   *vecptr++ = cfptr->CacheFrequentSeconds;

   /* timeouts */
   *vecptr++ = cfptr->TimeoutMinutesInput;
   *vecptr++ = cfptr->TimeoutMinutesOutput;
   *vecptr++ = cfptr->TimeoutMinutesNoProgress;
   *vecptr++ = cfptr->TimeoutSecondsKeepAlive;

   /* buffer sizes */
   if (cfptr == &Config)
   {
      /* in-use values */
      *vecptr++ = DclSysCommandSize;
      *vecptr++ = DclSysOutputSize;
      *vecptr++ = DclCgiPlusInSize;
      *vecptr++ = NetReadBufferSize;
      *vecptr++ = OutputBufferSize;
   }
   else
   {
      /* file values */
      *vecptr++ = cfptr->BufferSizeDclCommand;
      *vecptr++ = cfptr->BufferSizeDclOutput;
      *vecptr++ = cfptr->BufferSizeDclCgiPlusIn;
      *vecptr++ = cfptr->BufferSizeNetRead;
      *vecptr++ = cfptr->BufferSizeNetWrite;
   }

   status = sys$faol (&ReportGeneralFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      return;
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   /*****************************/
   /* welcome (home page) files */
   /*****************************/

   for (idx = 0; idx < cfptr->HomePageCount; idx++)
   {
      memcpy (sptr = Buffer, "<TR><TD><TT>", 12);
      sptr += 12;
      for (cptr = cfptr->HomePageArray[idx]; *cptr; *sptr++ = *cptr++);
      memcpy (sptr, "</TT></TD></TR>\n", 17);  /* includes terminating null */
      sptr += 17;
      NetWriteBuffered (rqptr, 0, Buffer, sptr-Buffer);
   }

   /*********************/
   /* host access, etc. */
   /*********************/

   vecptr = FaoVector;

   if (cfptr->AcceptHostsPtr == NULL)
      *vecptr++ = NotNone;
   else
      *vecptr++ = ConfigCommaList (rqptr, cfptr->AcceptHostsPtr, ',');

   if (cfptr->RejectHostsPtr == NULL)
      *vecptr++ = NotNone;
   else
      *vecptr++ = ConfigCommaList (rqptr, cfptr->RejectHostsPtr, ',');

   /* proxy serving */
   REPBOOL (cfptr->ProxyServingEnabled)
   REPBOOL (cfptr->ProxyCacheEnabled)
   REPBOOL (cfptr->ProxyAddForwardedByEnabled)
   REPBOOL (cfptr->ProxyReportLog)
   REPBOOL (cfptr->ProxyReportCacheLog)
   *vecptr++ = cfptr->ProxyHostLookupRetryCount;
   *vecptr++ = cfptr->ProxyHostCachePurgeHours;
   *vecptr++ = cfptr->ProxyCacheFileKBytesMax;
   if (cfptr->ProxyCacheReloadList[0])
      *vecptr++ = cfptr->ProxyCacheReloadList;
   else
      *vecptr++ = NotNone;
   *vecptr++ = cfptr->ProxyCacheRoutineHourOfDay;
   if (cfptr->ProxyCachePurgeList[0])
      *vecptr++ = cfptr->ProxyCachePurgeList;
   else
      *vecptr++ = NotNone;
   *vecptr++ = cfptr->ProxyCacheDeviceCheckMinutes;
   *vecptr++ = cfptr->ProxyCacheDeviceMaxPercent;
   *vecptr++ = cfptr->ProxyCacheDevicePurgePercent;

   /* authentication */
   *vecptr++ = cfptr->AuthCacheMinutes;
   *vecptr++ = cfptr->AuthRevalidateUserMinutes;
   *vecptr++ = cfptr->AuthFailureLimit;
   REPBOOL (cfptr->AuthBasicEnabled)
   REPBOOL (cfptr->AuthDigestEnabled)
   *vecptr++ = cfptr->AuthDigestNonceGetLifeTime;
   *vecptr++ = cfptr->AuthDigestNonceGetLifeTime;

   /* DCL/Scripting */
   REPBOOL (cfptr->DclSpawnAuthPriv)
   REPBOOL (cfptr->DclFullRequest)
   *vecptr++ = cfptr->DclSubprocessSoftLimit;
   *vecptr++ = cfptr->DclSubprocessHardLimit;
   *vecptr++ = cfptr->DclZombieLifeTime;
   *vecptr++ = cfptr->DclCgiPlusLifeTime;
   *vecptr++ = cfptr->DECnetReuseLifeTime;
   *vecptr++ = cfptr->DECnetConnectListMax;

   status = sys$faol (&ReportHostsFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      return;
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   /****************/
   /* script types */
   /****************/

   if (cfptr->ScriptRunTimeCount)
   {
      for (idx = 0; idx < cfptr->ScriptRunTimeCount; idx++)
      {
         sptr = Buffer;
         memcpy (sptr, "<TR><TD></TD><TD><TT>", 21);
         sptr += 21;
         for (cptr = cfptr->ScriptRunTimeArray[idx];
              *cptr && *cptr != ';';
              *sptr++ = *cptr++);
         memcpy (sptr, "</TT></TD><TD><TT>", 18);
         sptr += 18;
         if (*cptr) cptr++;
         if (*cptr) cptr++;
         if (*cptr == '@' || *cptr == '$')
         {
            while (*cptr && *cptr != '\"') cptr++;
            if (*cptr) cptr++;
            while (*cptr)
            {
               /* unescape any quotes */
               if (cptr[0] == '\"' && cptr[1] == '\"')
                  cptr++;
               else
               if (*cptr == '\"')
                  break;
               *sptr++ = *cptr++;
            }
         }
         else
            while (*cptr) *sptr++ = *cptr++;
         memcpy (sptr, "</TT></TD></TR>\n", 17);
         sptr += 16;
         NetWriteBuffered (rqptr, 0, Buffer, sptr-Buffer);
      }
   }
   else
   {
      static char  ScriptRunTimeNone [] =
"<TR><TD COLSPAN=2><FONT SIZE=-1><I>(none)</I></FONT></TD></TR>\n";

      NetWriteBuffered (rqptr, 0, ScriptRunTimeNone,
                                  sizeof(ScriptRunTimeNone)-1);
   }

   /*************/
   /* SSI, etc. */
   /*************/

   vecptr = FaoVector;

   /* SSI (.shtml) */
   REPBOOL (cfptr->SsiEnabled)
   REPBOOL (cfptr->SsiExecEnabled)
   REPBOOL (cfptr->SsiAccessesEnabled)

   REPBOOL (cfptr->DirAccess)
   if (cfptr->DefaultDirLayout[0])
      *vecptr++ = cfptr->DefaultDirLayout;
   else
      *vecptr++ = NotNone;
   *vecptr++ = cfptr->DirDescriptionLines;
   REPBOOL (cfptr->DirOwnerEnabled)
   REPBOOL (cfptr->DirPreExpired)
   REPBOOL (cfptr->DirWildcardEnabled)
   REPBOOL (cfptr->DirNoImpliedWildcard)
   REPBOOL (cfptr->DirNoPrivIgnore)
   REPBOOL (cfptr->DirReadMeTop | cfptr->DirReadMeBottom)
   if (cfptr->DirReadMeTop)
      *vecptr++ = "top";
   else
   if (cfptr->DirReadMeBottom)
      *vecptr++ = "bottom";
   else
      *vecptr++ = "";

   status = sys$faol (&ReportSsiFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      return;
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   /*********************/
   /* directory (contd) */
   /*********************/

   for (idx = 0; idx < cfptr->ReadMeFileCount; idx++)
   {
      sptr = Buffer;
      if (idx)
      {
         memcpy (sptr, "<TR><TD></TD>", 13);
         sptr += 13;
      }
      memcpy (sptr, "<TD><TT>", 8);
      sptr += 8;
      for (cptr = cfptr->ReadMeFileArray[idx]; *cptr; *sptr++ = *cptr++);
      memcpy (sptr, "</TT></TD></TR>\n", 17);  /* includes terminating null */
      sptr += 16;
      NetWriteBuffered (rqptr, 0, Buffer, sptr - Buffer);
   }

   NetWriteBuffered (rqptr, 0, ReportIconsFaoDsc.dsc$a_pointer,
                               ReportIconsFaoDsc.dsc$w_length);

   /*********/
   /* icons */
   /*********/

   ilptr = cfptr->IconListHeadPtr;
   while (ilptr != NULL)
   {
      /* find the ALT="" text from the icon IMG */ 
      IconAltText[0] = IconPath[0] = '\0';
      cptr = (char*)ilptr + cfptr->IconStructOverhead;
      while (*cptr && !ISLWS(*cptr)) cptr++;
      if (*cptr) cptr++;
      while (*cptr)
      {
         if (*cptr == 'S' && !memcmp (cptr, "SRC=\"", 5))
         {
            cptr += 5;
            sptr = IconPath;
            while (*cptr && *cptr != '\"') *sptr++ = *cptr++;
            *sptr = '\0';
         }
         else
         if (*cptr == 'A' && !memcmp (cptr, "ALT=\"", 5))
         {
            cptr += 5;
            sptr = IconAltText;
            while (*cptr && *cptr != '\"') *sptr++ = *cptr++;
            *sptr = '\0';
         }
         else
            cptr++;
      }
      if (!IconAltText[0])
         strcpy (IconAltText, NotNone);
      CopyToHtml (HtmlIconAltText, sizeof(HtmlIconAltText), IconAltText, -1);
      if (!IconPath[0])
         strcpy (IconPath, NotNone);

      memcpy (sptr = Buffer, "<TR><TD>", 8);
      sptr += 8;
      cptr = (char*)ilptr + cfptr->IconStructOverhead;
      while (*cptr && *cptr != ' ') *sptr++ = tolower(*cptr++);
      if (*cptr) cptr++;
      memcpy (sptr, "</TD><TD ALIGN=center>", 22);
      sptr += 22;
      while (*cptr) *sptr++ = *cptr++;
      memcpy (sptr, "</TD><TD>", 9);
      sptr += 9;
      for (cptr = IconPath; *cptr; *sptr++ = *cptr++);
      memcpy (sptr, "</TD><TD>", 9);
      sptr += 9;
      for (cptr = HtmlIconAltText; *cptr; *sptr++ = *cptr++);
      memcpy (sptr, "</TD></TR>\n", 12);  /* includes null */
      sptr += 11;

      NetWriteBuffered (rqptr, 0, Buffer, sptr - Buffer);

      ilptr = ilptr->NextPtr;
   }

   /*****************/
   /* content-types */
   /*****************/

   NetWriteBuffered (rqptr, 0, ReportContentTypeFaoDsc.dsc$a_pointer,
                              ReportContentTypeFaoDsc.dsc$w_length);

   clptr = cfptr->ContentTypeListHeadPtr;
   for (;;)
   {
      if (clptr == NULL) break;

      if (clptr->AutoScriptNamePtr[0] == '/')
         AutoScriptNamePtr = clptr->AutoScriptNamePtr;
      else
         AutoScriptNamePtr = "";

      if (clptr->ContentTypeUnknown)
         ContentDescriptionPtr = cfptr->ContentTypeDefaultPtr;
      else
         ContentDescriptionPtr = clptr->ContentDescriptionPtr;

      status = sys$fao (&ContentTypeFaoDsc, &Length, &BufferDsc,
                        clptr->SuffixPtr,
                        clptr->ContentTypePtr,
                        clptr->IconPtr,
                        AutoScriptNamePtr,
                        ContentDescriptionPtr);
      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         rqptr->ErrorTextPtr = "sys$fao()";
         ErrorVmsStatus (rqptr, status, FI_LI);
         return;
      }

      Buffer[Length] = '\0';
      NetWriteBuffered (rqptr, 0, Buffer, Length);

      clptr = clptr->NextPtr;
   }

   /**************/
   /* end report */
   /**************/

   NetWriteBuffered (rqptr, 0, EndPage, sizeof(EndPage)-1);
}

/*****************************************************************************/
/*
This function just wraps the revision function, loading a temporary database
if necessary for reporting from the configuration file.
*/ 

ConfigRevise
(
struct RequestStruct *rqptr,
void *NextTaskFunction,
boolean  UseServerDatabase
)
{
   int  status;
   struct ConfigStruct  LocalConfig;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigRevise()\n");

   if (UseServerDatabase)
      ConfigReviseNow (rqptr, &Config);
   else
   {
      memset (&LocalConfig, 0, sizeof(struct ConfigStruct));
      /* indicate it's being used for a report */
      Config.RequestPtr = rqptr;
      Configure (&LocalConfig);
      ConfigReviseNow (rqptr, &LocalConfig);
      ConfigFree (&LocalConfig);
   }

   NetWriteBuffered (rqptr, NextTaskFunction, NULL, 0);
}

/*****************************************************************************/
/*
Return a form for the HTTPd server's configuration.  POSTing the form to the
server results in the URL-decoded form as a configuration file.  This function
blocks while executing.
*/ 

ConfigReviseNow
(
struct RequestStruct *rqptr,
struct ConfigStruct *cfptr
)
{
#define SIZEOF_SPECIAL_ICON 256

#define REPBOOL(b)\
if (b)\
{\
   *vecptr++ = RadioButtonChecked;\
   *vecptr++ = RadioButtonUnchecked;\
}\
else\
{\
   *vecptr++ = RadioButtonUnchecked;\
   *vecptr++ = RadioButtonChecked;\
}

   static $DESCRIPTOR (ResponseFaoDsc,
"<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>HTTPd !AZ ... Configuration</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1><NOBR>HTTPd !AZ</NOBR></H1>\n\
<H2>Revise Configuration</H2>\n");

   static $DESCRIPTOR (ProblemReportFaoDsc,
"<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH><FONT COLOR=\"#ff0000\">!UL Report!AZ !AZ</FONT></TH></TR>\n\
<TR><TD><PRE>!AZ</PRE></TD></TR>\n\
</TABLE>\n");

   static $DESCRIPTOR (ConfigGeneralFaoDsc,
"<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Source: &quot;!AZ&quot;</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=3 CELLSPACING=0 BORDER=0>\n\
<TR>\
<TH ALIGN=right>File:</TH>\
<TD ALIGN=left>!AZ</TD>\
<TD>&nbsp;</TD><TH>[<A HREF=\"!AZ\">View</A>]</TH>\
<TH>[<A HREF=\"!AZ!AZ\">Edit</A>]</TH>\
</TR>\n\
<TR><TH ALIGN=right>!AZ</TH>\
<TD ALIGN=left>!AZ</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<FORM METHOD=POST ACTION=\"!AZ\">\n\
\
<INPUT TYPE=hidden NAME=x VALUE=\"\
\# Configuration:  !AZ&#10;\
\#                 !AZ&#10;\
\# Last Modified:  !AZ&#10;\
\#                 !AZ@!AZ&#10;\
\">\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Service</TH></TR>\n\
<TR><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[Service]&#10;\">\n\
<TEXTAREA NAME=Service ROWS=5 COLS=60 NOWRAP>\
!AZ\
</TEXTAREA>\n\
<BR><NOBR><B>Service Not Found URL:</NOBR></B>&nbsp;\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[ServiceNotFoundURL]  \">\n\
<INPUT TYPE=text SIZE=40 NAME=ServiceNotFoundURL VALUE=!AZ>\n\
</TD></TR>\n\
</TABLE>\n\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>General</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=3 CELLSPACING=0 BORDER=0>\n\
<TR><TH ALIGN=right>Busy:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[Busy]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=Busy VALUE=!UL>\n\
<FONT SIZE=-1>connections</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Request History:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[RequestHistory]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=RequestHistory VALUE=!UL>\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Activity History:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[ActivityDays]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=ActivityDays VALUE=!UL>\n\
<FONT SIZE=-1>days</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Charset Default:</TH><TD COLSPAN=2>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[CharsetDefault]  \">\n\
<INPUT TYPE=text size=40 NAME=Search VALUE=\"!AZ\">\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Comments:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[CommentedInfo]  \">\n\
<INPUT TYPE=radio NAME=CommentedInfo VALUE=enabled!AZ>yes\n\
<INPUT TYPE=radio NAME=CommentedInfo VALUE=disabled!AZ>no\n\
</TD></TR>\n\
<TR><TH ALIGN=right>DNS Lookup:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DNSLookup]  \">\n\
<INPUT TYPE=radio NAME=DNSLookup VALUE=enabled!AZ>yes\n\
<INPUT TYPE=radio NAME=DNSLookup VALUE=disabled!AZ>no\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Error Report Path:</TH><TD COLSPAN=2>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[ErrorReportPath]  \">\n\
<INPUT TYPE=text size=40 NAME=ErrorReportPath VALUE=\"!AZ\">\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Error Recommend:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[ErrorRecommend]  \">\n\
<INPUT TYPE=radio NAME=ErrorRecommend VALUE=enabled!AZ>yes\n\
<INPUT TYPE=radio NAME=ErrorRecommend VALUE=disabled!AZ>no\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Error Source Information:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[ErrorSourceInfo]  \">\n\
<INPUT TYPE=radio NAME=ErrorSourceInfo VALUE=enabled!AZ>yes\n\
<INPUT TYPE=radio NAME=ErrorSourceInfo VALUE=disabled!AZ>no\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Monitor:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[Monitor]  \">\n\
<INPUT TYPE=radio NAME=Monitor VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=Monitor VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH ALIGN=right>IP Port:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[Port]  \">\n\
<INPUT TYPE=text SIZE=5 NAME=Port VALUE=!UL>\n\
<FONT SIZE=-1>(overridden by <I>service</I> directive)</FONT></TD></TR>\n\
<TR><TH ALIGN=right>StreamLF:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[StreamLF]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=StreamLF VALUE=!UL>\n\
<FONT SIZE=-1>kBytes</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Search Script:</TH><TD COLSPAN=2>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[SearchScript]  \">\n\
<INPUT TYPE=text size=40 NAME=SearchScript VALUE=\"!AZ\">\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Put/Post Max Length:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[PutMaxKbytes]  \">\n\
<INPUT TYPE=text SIZE=5 NAME=PutMaxKbytes VALUE=!UL>\n\
<FONT SIZE=-1>kBytes</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Put/Post Files:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[PutVersionLimit]  \">\n\
<INPUT TYPE=text SIZE=5 NAME=PutVersionLimit VALUE=!UL>\n\
<FONT SIZE=-1>version limit</FONT></TD></TR>\n\
</TD></TR>\n\
<TR><TH ALIGN=right>/~username/ Cache Entries:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[MapUserNameCacheEntries]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=MapUserNameCacheEntries VALUE=!UL>\n\
<FONT SIZE=-1>user names</FONT>\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Log</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=3 CELLSPACING=0 BORDER=0>\n\
<TR><TH ALIGN=right>Logging:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[Logging]  \">\n\
<INPUT TYPE=radio NAME=Logging VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=Logging VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Format:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[LogFormat]  \">\n\
<INPUT TYPE=text SIZE=10 NAME=LogFormat VALUE=!AZ></TD></TR>\n\
<TR><TH ALIGN=right>Naming:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[LogNaming]  \">\n\
<INPUT TYPE=text SIZE=10 NAME=LogNaming VALUE=!AZ></TD></TR>\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Period:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[LogPeriod]  \">\n\
<INPUT TYPE=text SIZE=10 NAME=LogPeriod VALUE=!AZ></TD></TR>\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Per-Service:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[LogPerService]  \">\n\
<INPUT TYPE=radio NAME=LogPerService VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=LogPerService VALUE=disabled!AZ>disabled\n\
<TR><TH ALIGN=right>File:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[LogFile]  \">\n\
<INPUT TYPE=text SIZE=40 NAME=LogFile VALUE=!AZ></TD></TR>\n\
<TR><TH ALIGN=right>Exclude Hosts:</TH></TR>\n\
<TR><TD COLSPAN=2>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[LogExcludeHosts]&#10;\">\n\
<TEXTAREA NAME=LogExcludeHosts ROWS=5 COLS=40 NOWRAP>\
!AZ\
</TEXTAREA>\n\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Cache</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=3 CELLSPACING=0 BORDER=0>\n\
<TR><TH ALIGN=right>Caching:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[Cache]  \">\n\
<INPUT TYPE=radio NAME=CacheEnabled VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=CacheEnabled VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Hash Table:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[CacheHashTableEntries]  \">\n\
<INPUT TYPE=text SIZE=4 NAME=CacheHashTableEntries VALUE=!UL>\n\
<FONT SIZE=-1>entries</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Max Entries:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[CacheEntriesMax]  \">\n\
<INPUT TYPE=text SIZE=4 NAME=CacheEntriesMax VALUE=!UL>\n\
<FONT SIZE=-1>files</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Max Cache Size:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[CacheTotalKBytesMax]  \">\n\
<INPUT TYPE=text SIZE=4 NAME=CacheTotalKBytesMax VALUE=!UL>\n\
<FONT SIZE=-1>kBytes</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Max File Size:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[CacheFileKBytesMax]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=CacheFileKBytesMax VALUE=!UL>\n\
<FONT SIZE=-1>kBytes</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Chunk Size:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[CacheChunkKBytes]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=CacheChunkKBytes VALUE=!UL>\n\
<FONT SIZE=-1>kBytes</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Validate:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[CacheValidateSeconds]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=CacheValidateSeconds VALUE=!UL>\n\
<FONT SIZE=-1>seconds</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Frequent After:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[CacheFrequentHits]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=CacheFrequentHits VALUE=!UL> \
<FONT SIZE=-1>hits, last within</FONT>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[CacheFrequentSeconds]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=CacheFrequentSeconds VALUE=!UL>\n\
<FONT SIZE=-1>seconds</FONT></TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Timeouts</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=3 CELLSPACING=0 BORDER=0>\n\
<TR><TH ALIGN=right>Input:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[TimeoutInput]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=TimeoutInput VALUE=!UL>\n\
<FONT SIZE=-1>minutes</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Output:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[TimeoutOutput]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=TimeoutOutput VALUE=!UL>\n\
<FONT SIZE=-1>minutes</FONT></TD></TR>\n\
<TR><TH ALIGN=right>No-Progress:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[TimeoutNoProgress]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=TimeoutNoProgress VALUE=!UL>\n\
<FONT SIZE=-1>minutes</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Keep-Alive:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[TimeoutKeepAlive]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=TimeoutKeepAlive VALUE=!UL>\n\
<FONT SIZE=-1>seconds</FONT></TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Buffer Sizes</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=3 CELLSPACING=0 BORDER=0>\n\
<TR><TH ALIGN=right>DCL <FONT SIZE=-1>SYS$COMMAND</FONT>:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[BufferSizeDclCommand]  \">\n\
<INPUT TYPE=text SIZE=4 NAME=BufferSizeDclCommand VALUE=!UL>\n\
</TD></TR>\n\
<TR><TH ALIGN=right>DCL <FONT SIZE=-1>SYS$OUTPUT</FONT>:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[BufferSizeDclOutput]  \">\n\
<INPUT TYPE=text SIZE=4 NAME=BufferSizeDclOutput VALUE=!UL>\n\
</TD></TR>\n\
<TR><TH ALIGN=right>DCL <FONT SIZE=-1>CGIPLUSIN</FONT>:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[BufferSizeDclCgiPlusIn]  \">\n\
<INPUT TYPE=text SIZE=4 NAME=BufferSizeDclCgiPlusIn VALUE=!UL>\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Network Read:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[BufferSizeNetRead]  \">\n\
<INPUT TYPE=text SIZE=4 NAME=BufferSizeNetRead VALUE=!UL>\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Network Write:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[BufferSizeNetWrite]  \">\n\
<INPUT TYPE=text SIZE=4 NAME=BufferSizeNetWrite VALUE=!UL>\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Note ...</A></TH>\
<TD>adjust with caution!!</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Welcome (index/home pages)</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=3 CELLSPACING=0 BORDER=0>\n\
<TR><TD COLSPAN=2><FONT SIZE=-1>Format: \"file-name.file-type\" \
(e.g. HOME.HTML)</FONT></TD></TR>\n\
<TR><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[Welcome]&#10;\">\n\
<TEXTAREA NAME=Welcome ROWS=5 COLS=40 NOWRAP>");

   static $DESCRIPTOR (ConfigAcceptFaoDsc,
"</TEXTAREA>\n\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Host Access Control</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=3 CELLSPACING=0 BORDER=0>\n\
<TR><TH>Accept</TH></TR>\n\
<TR><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[Accept]&#10;\">\n\
<TEXTAREA NAME=Accept ROWS=5 COLS=40 NOWRAP>\
!AZ\
</TEXTAREA>\n\
</TD></TR>\n\
<TR><TH>Reject</TH></TR>\n\
<TR><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[Reject]&#10;\">\n\
<TEXTAREA NAME=Reject ROWS=5 COLS=40 NOWRAP>\
!AZ\
</TEXTAREA>\n\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Proxy Serving</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=3 CELLSPACING=0 BORDER=0>\n\
<TR><TH ALIGN=right>Serving:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[ProxyServing]  \">\n\
<INPUT TYPE=radio NAME=ProxyServing VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=ProxyServing VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Cache:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[ProxyCacheEnabled]  \">\n\
<INPUT TYPE=radio NAME=ProxyCacheEnabled VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=ProxyCacheEnabled VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Add &quot;Forwarded: by&quot;:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[ProxyAddForwardedBy]  \">\n\
<INPUT TYPE=radio NAME=ProxyAddForwardedBy VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=ProxyAddForwardedBy VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Report Events To Log:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[ProxyReportLog]  \">\n\
<INPUT TYPE=radio NAME=ProxyReportLog VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=ProxyReportLog VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Report Cache Events To Log:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[ProxyReportCacheLog]  \">\n\
<INPUT TYPE=radio NAME=ProxyReportCacheLog VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=ProxyReportCacheLog VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Host Name Lookup Retry:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[ProxyHostLookupRetryCount]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=ProxyHostLookupRetryCount VALUE=!UL>\n\
<FONT SIZE=-1>count</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Host Name Cache Purge:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[ProxyHostCachePurgeHours]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=ProxyHostCachePurgeHours VALUE=!UL>\n\
<FONT SIZE=-1>hours</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Maximum Load Size:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[ProxyCacheFileKBytesMax]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=ProxyCacheFileKBytesMax VALUE=!UL>\n\
<FONT SIZE=-1>kBytes</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Reload List:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[ProxyCacheReloadList]  \">\n\
<INPUT TYPE=text SIZE=30 NAME=ProxyCacheReloadList VALUE=\"!AZ\">\n\
<FONT SIZE=-1>hours</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Routine Check:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[ProxyCacheRoutineHourOfDay]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=ProxyCacheRoutineHourOfDay VALUE=!UL>\n\
<FONT SIZE=-1>hour-of-day (00-23)</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Purge List:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[ProxyCachePurgeList]  \">\n\
<INPUT TYPE=text SIZE=30 NAME=ProxyCachePurgeList VALUE=\"!AZ\">\n\
<FONT SIZE=-1>hours</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Device Usage Check:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[ProxyCacheDeviceCheckMinutes]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=ProxyCacheDeviceCheckMinutes VALUE=!UL>\n\
<FONT SIZE=-1>minutes</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Device Usage Max:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[ProxyCacheDeviceMaxPercent]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=ProxyCacheDeviceMaxPercent VALUE=!UL>\n\
<FONT SIZE=-1>percent</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Device Reduce By:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[ProxyCacheDevicePurgePercent]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=ProxyCacheDevicePurgePercent VALUE=!UL>\n\
<FONT SIZE=-1>percent</FONT></TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Authorization</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=3 CELLSPACING=0 BORDER=0>\n\
<TR><TH ALIGN=right>Cache Valid:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[AuthCacheMinutes]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=AuthCacheMinutes VALUE=!UL>\n\
<FONT SIZE=-1>minutes</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Revalidate User:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[AuthRevalidateUserMinutes]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=AuthRevalidateUserMinutes VALUE=!UL>\n\
<FONT SIZE=-1>minutes</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Failure Limit:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[AuthFailureLimit]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=AuthFailureLimit VALUE=!UL>\n\
<FONT SIZE=-1>attempts</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Basic Authorization:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[AuthBasic]  \">\n\
<INPUT TYPE=radio NAME=AuthBasic VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=AuthBasic VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Digest Authorization:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[AuthDigest]  \">\n\
<INPUT TYPE=radio NAME=AuthDigest VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=AuthDigest VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH ALIGN=right>GET nonce lifetime:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[AuthDigestGetLife]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=AuthDigestGetLife VALUE=!UL>\n\
<FONT SIZE=-1>minutes</FONT></TD>\n\
</TD></TR>\n\
<TR><TH ALIGN=right>PUT nonce lifetime:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[AuthDigestPutLife]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=AuthDigestPutLife VALUE=!UL>\n\
<FONT SIZE=-1>minutes</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Note ...</A></TH>\
<TD>paths must also be configured!!</TD></TR>\n\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Scripting &nbsp;(DCL & DECnet)</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=3 CELLSPACING=0 BORDER=0>\n\
<TR><TH ALIGN=right>Spawn Authorized Privileges:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DclSpawnAuthPriv]  \">\n\
<INPUT TYPE=radio NAME=DclSpawnAuthPriv VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=DclSpawnAuthPriv VALUE=disabled!AZ>disabled\n\
</TD><TD></TD></TR>\n\
<TR><TH ALIGN=right>Supply Header and Body:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DclFullRequest]  \">\n\
<INPUT TYPE=radio NAME=DclFullRequest VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=DclFullRequest VALUE=disabled!AZ>disabled\n\
</TD><TD></TD></TR>\n\
<TR><TH ALIGN=right>Soft Limit:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DclSoftLimit]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=DclSoftLimit VALUE=!UL>\n\
<FONT SIZE=-1>subprocesses</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Hard Limit:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DclHardLimit]  \">\n\
<INPUT TYPE=text SIZE=3 NAME=DclHardLimit VALUE=!UL>\n\
<FONT SIZE=-1>subprocesses</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Zombie Life-Time:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DclZombieLifeTime]  \">\n\
<INPUT TYPE=text SIZE=4 NAME=DclZombieLifeTime VALUE=!UL>\n\
<FONT SIZE=-1>minutes</FONT></TD></TR>\n\
<TR><TH ALIGN=right>CGIplus Lifetime:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DclCgiPlusLifeTime]  \">\n\
<INPUT TYPE=text SIZE=4 NAME=DclCgiPlusLifeTime VALUE=!UL>\n\
<FONT SIZE=-1>minutes</FONT></TD></TR>\n\
<TR><TH ALIGN=right>DECnet Reuse Life-Time:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DECnetReuseLifeTime]  \">\n\
<INPUT TYPE=text SIZE=4 NAME=DECnetReuseLifeTime VALUE=!UL>\n\
<FONT SIZE=-1>minutes</FONT></TD></TR>\n\
<TR><TH ALIGN=right>DECnet Connect List Limit:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DECnetConnectListMax]  \">\n\
<INPUT TYPE=text SIZE=4 NAME=DECnetConnectListMax VALUE=!UL>\n\
<FONT SIZE=-1>connections</FONT></TD></TR>\n\
<TR><TH ALIGN=left>Script Run-Time:</TH></TR>\n\
<TR><TD COLSPAN=2><FONT SIZE=-1>\
Format: &quot;file-type [SPACE] verb&quot;<BR>\
(e.g. &quot;.PL [SPACE] PERL&quot; or &quot;.PL [SPACE] $PERL_EXE:PERL&quot;)\
</FONT>\
</TD></TR>\n\
<TR><TD COLSPAN=2>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DclScriptRunTime]&#10;\">\n\
<TEXTAREA NAME=DclScriptRunTime ROWS=5 COLS=60 NOWRAP>");

   static $DESCRIPTOR (ConfigSsiFaoDsc,
"</TEXTAREA>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Server-Side Includes</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=3 CELLSPACING=0 BORDER=0>\n\
<TR><TH ALIGN=right>SSI:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[SSI]  \">\n\
<INPUT TYPE=radio NAME=SSI VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=SSI VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Exec:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[SSIexec]  \">\n\
<INPUT TYPE=radio NAME=SSIexec VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=SSIexec VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Access Counting:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[SSIaccesses]  \">\n\
<INPUT TYPE=radio NAME=SSIaccesses VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=SSIaccesses VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Directory (Index of...)</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=3 CELLSPACING=0 BORDER=0>\n\
<TR><TH ALIGN=right>Access:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DirAccess]  \">\n\
<INPUT TYPE=radio NAME=DirAccess VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=DirAccess VALUE=selective!AZ>selective\n\
<INPUT TYPE=radio NAME=DirAccess VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Layout:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DirLayout]  \">\n\
<INPUT TYPE=text size=30 NAME=DirLayout VALUE=\"!AZ\">\n\
</TD></TR>\n\
<TR><TH ALIGN=right>HTML Description:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DirDescriptionLines]  \">\n\
<INPUT TYPE=text SIZE=6 NAME=DirDescriptionLines VALUE=!UL>\n\
<FONT SIZE=-1>lines</FONT></TD></TR>\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Owner:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DirOwner]  \">\n\
<INPUT TYPE=radio NAME=DirOwner VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=DirOwner VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Pre-Expired:</TH><TD>\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DirPreExpired]  \">\n\
<INPUT TYPE=radio NAME=DirPreExpired VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=DirPreExpired VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Wildcard:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DirWildcard]  \">\n\
<INPUT TYPE=radio NAME=DirWildcard VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=DirWildcard VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH ALIGN=right>No Implied Wildcard:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DirNoImpliedWildcard]  \">\n\
<INPUT TYPE=radio NAME=DirNoImpliedWildcard VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=DirNoImpliedWildcard VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Prot Violations Ignored:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DirNoPrivIgnore]  \">\n\
<INPUT TYPE=radio NAME=DirNoPrivIgnore VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=DirNoPrivIgnore VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Readme:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DirReadme]  \">\n\
<INPUT TYPE=radio NAME=DirReadme VALUE=top!AZ>top\n\
<INPUT TYPE=radio NAME=DirReadme VALUE=bottom!AZ>bottom\n\
<INPUT TYPE=radio NAME=DirReadme VALUE=disabled!AZ>none\n\
</TD></TR>\n\
<TR><TH ALIGN=left>Readme Files:</TH></TR>\n\
<TR><TD COLSPAN=2><FONT SIZE=-1>Format: &quot;file-name.file-type&quot; \
(e.g. README.HTML)</FONT></TD></TR>\n\
<TR><TD COLSPAN=2>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[DirReadmeFile]&#10;\">\n\
<TEXTAREA NAME=DirReadmeFile ROWS=5 COLS=40 NOWRAP>");

   static $DESCRIPTOR (ConfigIconsFaoDsc,
"</TEXTAREA>\n\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Icons</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=3 CELLSPACING=0 BORDER=0>\n\
<TR><TD COLSPAN=2><FONT SIZE=-1>Format: &quot;icon-path [SPACE] \
alternative-text [SPACE] content-type&quot;</FONT></TD></TR>\n\
<TR><TD COLSPAN=2>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[AddIcon]&#10;\">\n\
<TEXTAREA NAME=AddIcon ROWS=5 COLS=60 NOWRAP>");

   static $DESCRIPTOR (ConfigSpecialIconsFaoDsc,
"</TEXTAREA>\n\
</TD></TR>\n\
<TR><TH COLSPAN=2>Special Icons</TH></TR>\n\
<TR><TD COLSPAN=2><FONT SIZE=-1>Format: &quot;icon-path [SPACE] \
alternative-text&quot;</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Blank:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[AddBlankIcon]  \">\n\
<INPUT TYPE=text SIZE=40 NAME=AddBlankIcon VALUE=\"!AZ\">\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Directory:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[AddDirIcon]  \">\n\
<INPUT TYPE=text SIZE=40 NAME=AddDirIcon VALUE=\"!AZ\">\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Parent Directory:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[AddParentIcon]  \">\n\
<INPUT TYPE=text SIZE=40 NAME=AddParentIcon VALUE=\"!AZ\">\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Unknown:</TH><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[AddUnknownIcon]  \">\n\
<INPUT TYPE=text SIZE=40 NAME=AddUnknownIcon VALUE=\"!AZ\">\n\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n");

   static $DESCRIPTOR (ConfigContentTypesFaoDsc,
"<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Content Types</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=3 CELLSPACING=0 BORDER=0>\n\
<TR><TD COLSPAN=2><FONT SIZE=-1>Format: &quot;file-suffix [SPACE] \
content-type[;charset=] [SPACE] auto-script-path <B>or</B> a hyphen [SPACE] \
description (to-end-of-line)&quot;</FONT></TD></TR>\n\
<TR><TD>\n\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;[AddType]&#10;\">\n\
<TEXTAREA NAME=AddType ROWS=5 COLS=60 NOWRAP>");

   static $DESCRIPTOR (EndPageFaoDsc,
"</TEXTAREA>\n\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<INPUT TYPE=hidden NAME=x VALUE=\"&#10;# End!&#10;\">\n\
<P><INPUT TYPE=submit VALUE=\" Commit Changes \">\n\
<INPUT TYPE=reset VALUE=\" Reset \">\n\
</FORM>\n\
</BODY>\n\
</HTML>\n");

   static char  RadioButtonChecked [] = " CHECKED",
                RadioButtonUnchecked [] = "";

   register int  idx;
   register unsigned long  *vecptr;
   register char  *cptr, *sptr, *zptr;
   register struct ContentTypeStruct  *clptr;
   register struct IconStruct  *ilptr;

   int  status;
   unsigned short  Length;
   unsigned long  FaoVector [128];
   char  *AutoScriptNamePtr,
         *ContentDescriptionPtr,
         *IconAltPtr,
         *IconAltEndPositionPtr,
         *IconAltStartPositionPtr,
         *IconUrlPtr,
         *IconTypePtr,
         *IconTypePositionPtr,
         *SpecialIconPtr;
   char  Buffer [16384],
         HtmlIconUrl [256],
         Scratch [256],
         SpecialIconBlank [SIZEOF_SPECIAL_ICON],
         SpecialIconDir [SIZEOF_SPECIAL_ICON],
         SpecialIconParent [SIZEOF_SPECIAL_ICON],
         SpecialIconUnknown [SIZEOF_SPECIAL_ICON],
         TimeCurrent [32];
   struct ContentTypeStruct  ContentType;
   struct ConfigStruct  LocalConfig;
   $DESCRIPTOR (BufferDsc, Buffer);
   $DESCRIPTOR (ScratchDsc, Scratch);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigReviseNow()\n");

   rqptr->ResponsePreExpired = PRE_EXPIRE_ADMIN;
   HTTP_HEADER_200_HTML (rqptr);

   vecptr = FaoVector;

   *vecptr++ = HtmlMetaInfo (rqptr, NULL);
   *vecptr++ = ServerHostPort;
   *vecptr++ = ServerHostPort;

   status = sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      return;
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   if (cfptr->ProblemReportLength)
   {
      vecptr = FaoVector;
      if ((*vecptr++ = cfptr->ProblemCount) == 1)
         *vecptr++ = "";
      else
         *vecptr++ = "s";
      if (cfptr == &Config)
         *vecptr++ = "At Startup";
      else
         *vecptr++ = "During Load";
      *vecptr++ = cfptr->ProblemReportPtr;

      status = sys$faol (&ProblemReportFaoDsc, &Length, &BufferDsc,
                         &FaoVector);
      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         rqptr->ErrorTextPtr = "sys$faol()";
         ErrorVmsStatus (rqptr, status, FI_LI);
         return;
      }

      Buffer[Length] = '\0';
      NetWriteBuffered (rqptr, 0, Buffer, Length);
   }

   strcpy (TimeCurrent, DayDateTime (&rqptr->BinaryTime, 20));
   cptr = MapVmsPath (cfptr->LoadFileName, rqptr);

   vecptr = FaoVector;

   /* source information (i.e. from executing server or static file) */
   if (cfptr == &Config)
      *vecptr++ = "Server";
   else
      *vecptr++ = "File";
   *vecptr++ = cfptr->LoadFileName;
   *vecptr++ = cptr;
   *vecptr++ = INTERNAL_SCRIPT_UPD;
   *vecptr++ = cptr;

   if (cfptr == &Config)
   {
      *vecptr++ = "Loaded:";
      *vecptr++ = DayDateTime (&cfptr->LoadBinTime, 20);
   }
   else
   {
      *vecptr++ = "Revised:";
      *vecptr++ = DayDateTime (&cfptr->RevBinTime, 20);
   }

   /* form action */
   *vecptr++ = cptr;

   /* comments */
   *vecptr++ = ServerHostPort;
   *vecptr++ = SoftwareID;
   *vecptr++ = TimeCurrent;
   *vecptr++ = rqptr->RemoteUser;
   *vecptr++ = rqptr->ClientHostName;

   /* services */
   *vecptr++ = ConfigCommaList (rqptr, cfptr->ServicePtr, '\n');
   *vecptr++ = cfptr->ServiceNotFoundUrl;

   /* general */
   if (cfptr->Busy)
      *vecptr++ = cfptr->Busy;
   else
      *vecptr++ = DEFAULT_CONCURRENT_CONNECT_MAX;
   *vecptr++ = cfptr->RequestHistory;
   *vecptr++ = cfptr->ActivityNumberOfDays;
   *vecptr++ = cfptr->CharsetDefault;
   REPBOOL (cfptr->IncludeCommentedInfo)
   REPBOOL (cfptr->DnsLookup)
   *vecptr++ = cfptr->ErrorReportPath;
   REPBOOL (cfptr->ErrorRecommend)
   REPBOOL (cfptr->ErrorSourceInfo)
   REPBOOL (cfptr->MonitorEnabled)
   *vecptr++ = cfptr->ServerPort;
   *vecptr++ = cfptr->StreamLfConversionMaxKbytes;
   *vecptr++ = cfptr->KeywordSearchScript;
   *vecptr++ = cfptr->PutMaxKbytes;
   *vecptr++ = cfptr->PutVersionLimit;
   if (cfptr == &Config)
      *vecptr++ = MapUrlUserNameCacheEntries;
   else
      *vecptr++ = cfptr->MapUserNameCacheEntries;

   /* logging */
   REPBOOL (cfptr->LoggingEnabled)
   *vecptr++ = cfptr->LogFormat;
   *vecptr++ = cfptr->LogNaming;
   *vecptr++ = cfptr->LogPeriod;
   REPBOOL (cfptr->LogPerService)
   *vecptr++ = cfptr->LogFileName;
   *vecptr++ = ConfigCommaList (rqptr, cfptr->LogExcludeHostsPtr, '\n');

   /* cache */
   REPBOOL (cfptr->CacheEnabled)
   *vecptr++ = cfptr->CacheHashTableEntries;
   *vecptr++ = cfptr->CacheEntriesMax;
   *vecptr++ = cfptr->CacheTotalKBytesMax;
   *vecptr++ = cfptr->CacheFileKBytesMax;
   *vecptr++ = cfptr->CacheChunkKBytes;
   *vecptr++ = cfptr->CacheValidateSeconds;
   *vecptr++ = cfptr->CacheFrequentHits;
   *vecptr++ = cfptr->CacheFrequentSeconds;

   /* timeouts */
   *vecptr++ = cfptr->TimeoutMinutesInput;
   *vecptr++ = cfptr->TimeoutMinutesOutput;
   *vecptr++ = cfptr->TimeoutMinutesNoProgress;
   *vecptr++ = cfptr->TimeoutSecondsKeepAlive;

   /* buffer sizes */
   if (cfptr == &Config)
   {
      /* in-use values */
      *vecptr++ = DclSysCommandSize;
      *vecptr++ = DclSysOutputSize;
      *vecptr++ = DclCgiPlusInSize;
      *vecptr++ = NetReadBufferSize;
      *vecptr++ = OutputBufferSize;
   }
   else
   {
      /* file values */
      *vecptr++ = cfptr->BufferSizeDclCommand;
      *vecptr++ = cfptr->BufferSizeDclOutput;
      *vecptr++ = cfptr->BufferSizeDclCgiPlusIn;
      *vecptr++ = cfptr->BufferSizeNetRead;
      *vecptr++ = cfptr->BufferSizeNetWrite;
   }

   status = sys$faol (&ConfigGeneralFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      return;
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   /*****************************/
   /* welcome (home page) files */
   /*****************************/

   for (idx = 0; idx < cfptr->HomePageCount; idx++)
   {
      zptr = (sptr = Scratch) + sizeof(Scratch);
      if (!idx) *sptr++ = '\n';
      for (cptr = cfptr->HomePageArray[idx];
           *cptr && sptr < zptr;
           *sptr++ = *cptr++);
      if (sptr < zptr) *sptr++ = '\n';
      if (sptr >= zptr)
         strcpy (Scratch, ConfigStringOverflow);
      else
         *sptr = '\0';

      Length = CopyToHtml (Buffer, sizeof(Buffer), Scratch, -1);
      NetWriteBuffered (rqptr, 0, Buffer, Length);
   }

   /********/
   /* more */
   /********/

   vecptr = FaoVector;

   /* accept and reject hosts */
   *vecptr++ = ConfigCommaList (rqptr, cfptr->AcceptHostsPtr, '\n');
   *vecptr++ = ConfigCommaList (rqptr, cfptr->RejectHostsPtr, '\n');

   /* proxy serving */
   REPBOOL (cfptr->ProxyServingEnabled)
   REPBOOL (cfptr->ProxyCacheEnabled)
   REPBOOL (cfptr->ProxyAddForwardedByEnabled)
   REPBOOL (cfptr->ProxyReportLog)
   REPBOOL (cfptr->ProxyReportCacheLog)
   *vecptr++ = cfptr->ProxyHostLookupRetryCount;
   *vecptr++ = cfptr->ProxyHostCachePurgeHours;
   *vecptr++ = cfptr->ProxyCacheFileKBytesMax;
   *vecptr++ = cfptr->ProxyCacheReloadList;
   *vecptr++ = cfptr->ProxyCacheRoutineHourOfDay;
   *vecptr++ = cfptr->ProxyCachePurgeList;
   *vecptr++ = cfptr->ProxyCacheDeviceCheckMinutes;
   *vecptr++ = cfptr->ProxyCacheDeviceMaxPercent;
   *vecptr++ = cfptr->ProxyCacheDevicePurgePercent;

   /* authentication */
   *vecptr++ = cfptr->AuthCacheMinutes;
   *vecptr++ = cfptr->AuthRevalidateUserMinutes;
   *vecptr++ = cfptr->AuthFailureLimit;
   REPBOOL (cfptr->AuthBasicEnabled)
   REPBOOL (cfptr->AuthDigestEnabled)
   *vecptr++ = cfptr->AuthDigestNonceGetLifeTime;
   *vecptr++ = cfptr->AuthDigestNonceGetLifeTime;

   /* DCL/scripting */
   REPBOOL (cfptr->DclSpawnAuthPriv)
   REPBOOL (cfptr->DclFullRequest)
   *vecptr++ = cfptr->DclSubprocessSoftLimit;
   *vecptr++ = cfptr->DclSubprocessHardLimit;
   *vecptr++ = cfptr->DclZombieLifeTime;
   *vecptr++ = cfptr->DclCgiPlusLifeTime;
   *vecptr++ = cfptr->DECnetReuseLifeTime;
   *vecptr++ = cfptr->DECnetConnectListMax;

   status = sys$faol (&ConfigAcceptFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      return;
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   /****************/
   /* script types */
   /****************/

   for (idx = 0; idx < cfptr->ScriptRunTimeCount; idx++)
   {
      sptr = Buffer;
      for (cptr = cfptr->ScriptRunTimeArray[idx];
           *cptr && *cptr != ';';
           *sptr++ = *cptr++);
      *sptr++ = ' ';
      if (*cptr) cptr++;
      if (*cptr) cptr++;
      if (*cptr == '@' || *cptr == '$')
      {
         while (*cptr && *cptr != '\"') cptr++;
         if (*cptr) cptr++;
         while (*cptr)
         {
            /* unescape any quotes */
            if (cptr[0] == '\"' && cptr[1] == '\"')
               cptr++;
            else
            if (*cptr == '\"')
               break;
            *sptr++ = *cptr++;
         }
      }
      else
         while (*cptr) *sptr++ = *cptr++;
      *sptr++ = '\n';
      *sptr = '\0';
      NetWriteBuffered (rqptr, 0, Buffer, sptr-Buffer);
   }

   /********/
   /* more */
   /********/

   vecptr = FaoVector;

   /* SSI (.shtml) */
   REPBOOL (cfptr->SsiEnabled)
   REPBOOL (cfptr->SsiExecEnabled)
   REPBOOL (cfptr->SsiAccessesEnabled)

   /* directory */
   if (cfptr->DirAccess)
   {
      *vecptr++ = " CHECKED";
      *vecptr++ = "";
      *vecptr++ = "";
   }
   else
   if (cfptr->DirAccessSelective)
   {
      *vecptr++ = "";
      *vecptr++ = " CHECKED";
      *vecptr++ = "";
   }
   else
   {
      *vecptr++ = "";
      *vecptr++ = "";
      *vecptr++ = " CHECKED";
   }
   *vecptr++ = cfptr->DefaultDirLayout;
   *vecptr++ = cfptr->DirDescriptionLines;
   REPBOOL (cfptr->DirOwnerEnabled)
   REPBOOL (cfptr->DirPreExpired)
   REPBOOL (cfptr->DirWildcardEnabled)
   REPBOOL (cfptr->DirNoImpliedWildcard)
   REPBOOL (cfptr->DirNoPrivIgnore)
   if (cfptr->DirReadMeTop)
   {
      *vecptr++ = " CHECKED";
      *vecptr++ = "";
      *vecptr++ = "";
   }
   else
   if (cfptr->DirReadMeBottom)
   {
      *vecptr++ = "";
      *vecptr++ = " CHECKED";
      *vecptr++ = "";
   }
   else
   {
      *vecptr++ = "";
      *vecptr++ = "";
      *vecptr++ = " CHECKED";
   }
   REPBOOL (cfptr->DirWildcardEnabled)

   status = sys$faol (&ConfigSsiFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      return;
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   /****************/
   /* readme files */
   /****************/

   for (idx = 0; idx < cfptr->ReadMeFileCount; idx++)
   {
      if (!idx) NetWriteBuffered (rqptr, 0, "\n", 1);
      Length = CopyToHtml (Buffer, sizeof(Buffer)-1,
                           cfptr->ReadMeFileArray[idx], -1);
      Buffer[Length++] = '\n';
      Buffer[Length] = '\0';
      NetWriteBuffered (rqptr, 0, Buffer, Length);
   }

   /********/
   /* more */
   /********/

   NetWriteBuffered (rqptr, 0, ConfigIconsFaoDsc.dsc$a_pointer,
                              ConfigIconsFaoDsc.dsc$w_length);

   /*********/
   /* icons */
   /*********/

   SpecialIconBlank[0] = SpecialIconDir[0] = SpecialIconParent[0] =
      SpecialIconUnknown[0] = '\n';
   SpecialIconBlank[1] = SpecialIconDir[1] = SpecialIconParent[1] =
      SpecialIconUnknown[1] = '\0';

   ilptr = cfptr->IconListHeadPtr;
   while (ilptr != NULL)
   {
      /*
         Convert from 1 back to 2.

         1.  sprintf ((char*)ilptr + cfptr->IconStructOverhead,
                      "%s <IMG ALIGN=top SRC=\"%s\" ALT=\"%s\">",
                      ContentTypePtr, IconUrlPtr, AltTextPtr);

         2.  <URL>[SPACE]<alternative-text>[SPACE]<content-type>
      */

      IconTypePtr = cptr = (char*)ilptr + cfptr->IconStructOverhead;
      while (*cptr && !ISLWS(*cptr)) cptr++;
      if (*cptr) cptr++;
      while (*cptr)
      {
         if (*cptr == 'S' && !memcmp (cptr, "SRC=\"", 5))
         {
            cptr += 5;
            IconUrlPtr = cptr;
            while (*cptr && *cptr != '\"') cptr++;
            if (*cptr) cptr++;
         }
         else
         if (*cptr == 'A' && !memcmp (cptr, "ALT=\"", 5))
         {
            cptr += 5;
            IconAltPtr = cptr;
            while (*cptr && *cptr != '\"') cptr++;
            if (*cptr) cptr++;
         }
         else
            cptr++;
      }
      zptr = (sptr = Scratch) + sizeof(Scratch);
      if (ilptr == cfptr->IconListHeadPtr) *sptr++ = '\n';
      /* icon URL */
      for (cptr = IconUrlPtr;
           *cptr && *cptr != '\"' && sptr < zptr;
           *sptr++ = *cptr++);
      if (sptr < zptr) *sptr++ = ' ';
      if (sptr < zptr) *sptr++ = ' ';
      /* note the start of alternative text */
      IconAltStartPositionPtr = sptr;
      /* alternative text */
      for (cptr = IconAltPtr; *cptr && *cptr != '\"' && sptr < zptr; cptr++)
      {
         if (*cptr == ' ')
            *sptr++ = '_';
         else
            *sptr++ = *cptr;
      }
      IconAltEndPositionPtr = sptr;
      if (sptr < zptr) *sptr++ = ' ';
      if (sptr < zptr) *sptr++ = ' ';
      /* note position of null-terminated content type in the scratch string */
      IconTypePositionPtr = sptr;
      /* get the content type */
      for (cptr = IconTypePtr;
           *cptr && *cptr != ' ' && sptr < zptr;
           *sptr++ = tolower(*cptr++));
      if (sptr >= zptr)
         Scratch[0] = '\0';
      else
         *sptr = '\0';

      SpecialIconPtr = NULL;
      if (strsame (IconTypePositionPtr, ConfigContentTypeBlank, -1))
      {
         /* replace the blanks with equivalent underscores */
         while (IconAltStartPositionPtr < IconAltEndPositionPtr)
            *IconAltStartPositionPtr++ = '_';
         SpecialIconPtr = SpecialIconBlank;
      }
      else
      if (strsame (IconTypePositionPtr, ConfigContentTypeDir, -1))
         SpecialIconPtr = SpecialIconDir;
      else
      if (strsame (IconTypePositionPtr, ConfigContentTypeParent, -1))
         SpecialIconPtr = SpecialIconParent;
      else
      if (strsame (IconTypePositionPtr, ConfigContentTypeUnknown, -1))
         SpecialIconPtr = SpecialIconUnknown;
      /* do not include content-type in special icon strings */
      if (SpecialIconPtr != NULL) *IconAltEndPositionPtr = '\0';

      if (sptr < zptr) *sptr++ = '\n';
      if (sptr >= zptr)
         strcpy (Scratch, ConfigStringOverflow);
      else
         *sptr = '\0';

      Length = CopyToHtml (Buffer, sizeof(Buffer), Scratch, -1);

      if (SpecialIconPtr == NULL)
         NetWriteBuffered (rqptr, 0, Buffer, Length);
      else
      {
         zptr = (sptr = SpecialIconPtr) + SIZEOF_SPECIAL_ICON;
         for (cptr = Buffer; *cptr && sptr < zptr; *sptr++ = *cptr++);
         if (sptr >= zptr)
            strcpy (SpecialIconPtr, ConfigStringOverflow);
         else
            *sptr = '\0';
      }

      ilptr = ilptr->NextPtr;
   }

   vecptr = FaoVector;

   *vecptr++ = SpecialIconBlank;
   *vecptr++ = SpecialIconDir;
   *vecptr++ = SpecialIconParent;
   *vecptr++ = SpecialIconUnknown;

   status = sys$faol (&ConfigSpecialIconsFaoDsc, &Length, &BufferDsc,
                      &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      return;
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   /*****************/
   /* content types */
   /*****************/

   NetWriteBuffered (rqptr, 0, ConfigContentTypesFaoDsc.dsc$a_pointer,
                              ConfigContentTypesFaoDsc.dsc$w_length);

   clptr = cfptr->ContentTypeListHeadPtr;
   for (;;)
   {
      if (clptr == NULL) break;

      if (clptr->AutoScriptNamePtr[0] == '/')
         AutoScriptNamePtr = clptr->AutoScriptNamePtr;
      else
         AutoScriptNamePtr = "-";

      if (clptr->ContentTypeUnknown)
         ContentDescriptionPtr = cfptr->ContentTypeDefaultPtr;
      else
         ContentDescriptionPtr = clptr->ContentDescriptionPtr;

      zptr = (sptr = Scratch) + sizeof(Scratch);
      if (clptr == cfptr->ContentTypeListHeadPtr) *sptr++ = '\n';
      for (cptr = clptr->SuffixPtr;
           *cptr && sptr < zptr;
           *sptr++ = *cptr++);
      if (sptr < zptr) *sptr++ = ' ';
      if (sptr < zptr) *sptr++ = ' ';
      if (strchr (clptr->ContentTypePtr, ';') == NULL)
      {
         for (cptr = clptr->ContentTypePtr;
              *cptr && sptr < zptr;
              *sptr++ = *cptr++);
      }
      else
      {
         if (sptr < zptr) *sptr++ = '\"';
         for (cptr = clptr->ContentTypePtr;
              *cptr && sptr < zptr;
              *sptr++ = *cptr++);
         if (sptr < zptr) *sptr++ = '\"';
      }
      if (sptr < zptr) *sptr++ = ' ';
      if (sptr < zptr) *sptr++ = ' ';
      for (cptr = AutoScriptNamePtr;
           *cptr && sptr < zptr;
           *sptr++ = *cptr++);
      if (sptr < zptr) *sptr++ = ' ';
      if (sptr < zptr) *sptr++ = ' ';
      for (cptr = ContentDescriptionPtr;
           *cptr && sptr < zptr;
           *sptr++ = *cptr++);
      if (sptr < zptr) *sptr++ = '\n';
      if (sptr >= zptr)
         strcpy (Scratch, ConfigStringOverflow);
      else
         *sptr = '\0';

      Length = CopyToHtml (Buffer, sizeof(Buffer), Scratch, -1);
      NetWriteBuffered (rqptr, 0, Buffer, Length);

      clptr = clptr->NextPtr;
   }

   /************/
   /* end form */
   /************/

   NetWriteBuffered (rqptr, 0, EndPageFaoDsc.dsc$a_pointer,
                               EndPageFaoDsc.dsc$w_length);
}

/*****************************************************************************/
/*
Takes a comma-separated list of elements and returns a pointer to a dynamically
allocated string with that list reformated for inclusion in HTML.  The
separator is intended to be either a newline or comma character.  If a comma an
extra space is added after each (so that when displayed in HTML it will be
wrapped).
*/

char* ConfigCommaList
(
struct RequestStruct *rqptr,
char *ListPtr,
char Separator
)
{
   register char  *cptr, *sptr, *zptr;

   int  Length;
   char  ListBuffer [8192],
         HtmlListBuffer [8192];

   /*********/
   /* begin */
   /*********/
   
   if (Debug) fprintf (stdout, "ConfigCommaList() |%s|\n", ListPtr);

   if (ListPtr == NULL) return ("");
   zptr = (sptr = ListBuffer) + sizeof(ListBuffer);
   for (cptr = ListPtr; *cptr && sptr < zptr; cptr++)
   {
      if (*cptr == ',')
          *sptr++ = Separator;
      else
          *sptr++ = *cptr;
      if (Separator == ',' && *cptr == ',' && sptr < zptr) *sptr++ = ' ';
   }
   if (sptr >= zptr) return (ConfigStringOverflow);
   *sptr = '\0';
   Length = CopyToHtml (HtmlListBuffer, sizeof(HtmlListBuffer), ListBuffer, -1);
   if (Length == -1) return (ConfigStringOverflow);
   sptr = VmGetHeap (rqptr, Length);
   memcpy (sptr, HtmlListBuffer, Length+1);
   if (Debug) fprintf (stdout, "|%s|\n", sptr);
   return (sptr);
}

/*****************************************************************************/
